"""
Elements
========
"""

from enum import Enum

import numpy as np
from ansys.dpf.core import field, nodes, property_field, scoping
from ansys.dpf.core.common import __write_enum_doc__, locations
from ansys.dpf.core.element_descriptor import ElementDescriptor
from ansys.dpf.core.errors import protect_grpc
from ansys.grpc.dpf import meshed_region_pb2


class Element:
    """Contains all properties of an element of a mesh.

    The element is created from the
    :class:`MeshedRegion <ansys.dpf.core.meshed_region.MeshedRegion>` class.
    Properties include the element ID, index, type, shape, and connectivity.

    Parameters
    ----------
    mesh : :class:`ansys.dpf.core.meshed_region.MeshedRegion`
        Mesh containing the element.
    elementid : int
        Number (ID) of the element.
    index : int
        Fortran-based (1-based) index of the element in the result.
    nodes : list
        List of DPF nodes belonging to the element.

    Examples
    --------
    Extract a single element from a meshed region.

    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> elements = model.metadata.meshed_region.elements
    >>> element = elements[0]

    List the coordinates belonging to the first node of the element.

    >>> element.nodes[0].coordinates
    [0.015, 0.045, 0.015]

    """

    def __init__(self, mesh, elementid, index, nodes):
        self._id = elementid
        self._index = index
        self._nodes = nodes
        self._mesh = mesh

    @property
    def node_ids(self):
        """IDs of all nodes in the element.

        Returns
        --------
        list
           List of IDs for all nodes in the element.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> element = elements[0]
        >>> element.node_ids
        [1, 26, 14, 12, 2, 27, 15, 13, 33, 64, 59, 30, 37, 65, 61, 34, 28, 81, 63, 58]

        """
        return [node.id for node in self._nodes]

    @property
    def id(self) -> int:
        """ID of the element.

        Returns
        -------
        int
            ID of the element.

        """
        return self._id

    @property
    def index(self) -> int:
        """Index of the element in the result.

        Returns
        -------
        int
            Index of the element in the result. This uses zero-based indexing starting at ``0``.

        """
        return self._index

    @property
    def nodes(self):
        """All nodes in the element.
        Returns
        --------
        list
            List of all nodes in the element.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> element = elements[0]
        >>> first_node = element.nodes[0]

        """
        return self._nodes

    @property
    def n_nodes(self) -> int:
        """Number of nodes in the element.

        Returns
        -------
        int
            Number of nodes.

        """
        return len(self._nodes)

    def __str__(self):
        txt = "DPF Element %d\n" % self.id
        txt += "\tIndex:      %7d\n" % self.index
        txt += "\tNodes:      %7d\n" % self.n_nodes
        txt += f"\tType:       {self.type}\n"
        txt += "\tShape:      %7s\n" % self.shape.capitalize()
        return txt

    @property
    def type(self) -> int:
        """Type of the element.

        Returns
        -------
        int
            Type of the element. For more information, see
            :class:`ansys.dpf.core.elements.element_types`.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> element = elements[0]
        >>> element.type
        <element_types.Hex20: 1>

        """
        return self._get_type()

    @protect_grpc
    def _get_type(self):
        """Retrieve the Ansys element type."""

        request = meshed_region_pb2.ElementalPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        request.index = self.index
        request.property = meshed_region_pb2.ELEMENT_TYPE
        return element_types(self._mesh._stub.GetElementalProperty(request).prop)

    @property
    def shape(self) -> str:
        """Shape of the element.

        Returns
        -------
        str
           Shape of the element, which can be ``"shell"``, ``"solid"``, ``"beam"``,
           or ``"unknown_shape"``.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> element = elements[0]
        >>> element.shape
        'solid'

        """
        return self._get_shape()

    @protect_grpc
    def _get_shape(self):
        """Retrieve the element shape."""
        request = meshed_region_pb2.ElementalPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        request.index = self.index
        request.property = meshed_region_pb2.ELEMENT_SHAPE
        prop = self._mesh._stub.GetElementalProperty(request).prop
        return meshed_region_pb2.ElementShape.Name(prop).lower()

    @property
    def connectivity(self):
        """Ordered list of node indices of the element.

        Returns
        --------
        list
            Ordered list of node indices.

        """
        list = []
        for node in self._nodes:
            list.append(node.index)
        return list


class Elements:
    """Contains elements belonging to a meshed region.

    Parameters
    ----------
    mesh : str
        Name of the meshed region.

    Examples
    --------
    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> elements = model.metadata.meshed_region.elements
    >>> elements.n_elements
    8

    """

    def __init__(self, mesh):
        self._mesh = mesh
        self._mapping_id_to_index = None

    def __str__(self):
        return "DPF Elements object with %d elements" % len(self)

    def __getitem__(self, index):
        """Retrieves element based on an index."""
        return self.element_by_index(index)

    def __len__(self):
        return self.n_elements

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def element_by_id(self, id) -> Element:
        """Retrieve an element by element ID.

        Parameters
        ----------
        id : int
            Number (ID) of the element.

        Returns
        -------
        Element
            Element object.

        """
        return self.__get_element(elementid=id)

    def element_by_index(self, index) -> Element:
        """Retrieve an element using its index.

        Parameters
        ----------
        index : int
            Zero-based index.

        Returns
        -------
        :class:`ansys.dpf.core.elements.ElementAdder`
            Yield element.

        Examples
        --------
        elements.element_by_index(0)

        Notes
        -----
        This is equivalent to ``elements[0]``

        """
        return self.__get_element(elementindex=index)

    def add_elements(self, num):
        """Add one or more elements in the mesh.

        Parameters
        ----------
        num : int
            Number of elements to add in the mesh.

        Returns
        -------
        :class:`ansys.dpf.core.elements.ElementAdder`
            Elements added to the mesh.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=3)
        >>> i=0
        >>> for node in meshed_region.nodes.add_nodes(4):
        ...     node.id = i+1
        ...     node.coordinates = [float(i), float(i), 0.0]
        ...     i=i+1
        >>> i=0
        >>> for element in meshed_region.elements.add_elements(3):
        ...     element.id=i+1
        ...     element.connectivity = [i, i+1]
        ...     element.is_beam=True #or is_solid, is_beam, is_point
        ...     i=i+1

        """
        request = meshed_region_pb2.AddRequest(mesh=self._mesh._message)
        for i in range(0, num):
            add = ElementAdder()
            yield add
            element_request = meshed_region_pb2.ElementRequest(id=add.id)
            element_request.connectivity.extend(add.connectivity)
            element_request.shape = meshed_region_pb2.ElementShape.Value(
                add.shape.upper()
            )
            request.elements.append(element_request)
        self._mesh._stub.Add(request)

    def add_solid_element(self, id, connectivity):
        """Add a solid 3D element in the mesh.

        Parameters
        ----------
        id : int
            ID to assign the new element.
        connectivity : list
            List of the node indices to connect to the new element.
        """
        self.add_element(id, "solid", connectivity)

    def add_shell_element(self, id, connectivity):
        """Add a shell 2D element in the mesh.

        Parameters
        ----------
        id : int
            ID to assign the new element.
        connectivity : list
            List of the node indices to connect to the new element.
        """
        self.add_element(id, "shell", connectivity)

    def add_beam_element(self, id, connectivity):
        """Add a beam 1D element in the mesh.

        Parameters
        ----------
        id : int
            ID to assign the new element.
        connectivity : list
            List of the node indices to connect to the new element.

        """
        self.add_element(id, "beam", connectivity)

    def add_point_element(self, id, connectivity):
        """Add a point element (one node connectivity) in the mesh.

        Parameters
        ----------
        id : int
            ID to assign the new element.
        connectivity : list
            List of the node indices to connect to the new element.

        """
        if not isinstance(connectivity, list):
            connectivity = [connectivity]
        self.add_element(id, "unknown_shape", connectivity)

    def add_element(self, id, shape, connectivity):
        """Add an element in the mesh.

        Parameters
        ----------
        id : int
            ID to assign the new element.
        shape : str
            Shape of the element. Options are ``"solid"``, ``"shell"``, ``"beam"``
            and ``"unknown_shape"``.
        connectivity : list
            List of the node indices to connect to the new element.

        """
        request = meshed_region_pb2.AddRequest(mesh=self._mesh._message)
        element_request = meshed_region_pb2.ElementRequest(id=id)
        element_request.connectivity.extend(connectivity)
        element_request.shape = meshed_region_pb2.ElementShape.Value(shape.upper())
        request.elements.extend([element_request])
        self._mesh._stub.Add(request)

    @protect_grpc
    def __get_element(self, elementindex=None, elementid=None):
        """Retrieve the element by ID or index.

        Parameters
        ----------
        elementid : int, optional
            ID of the element. The default is ``None``.
        elementindex : int, optional
            Index of the element. The default is ``None``.

        Returns
        -------
        element : Element
        """
        request = meshed_region_pb2.GetRequest()
        request.mesh.CopyFrom(self._mesh._message)
        if elementindex is None:
            request.id = elementid
        else:
            request.index = elementindex

        elementOut = self._mesh._stub.GetElement(request)
        nodesOut = []
        for node in elementOut.nodes:
            nodesOut.append(
                nodes.Node(self._mesh, node.id, node.index, node.coordinates)
            )
        return Element(self._mesh, elementOut.id, elementOut.index, nodesOut)

    @property
    def scoping(self) -> scoping.Scoping:
        """Scoping of the elements.

        Returns
        -------
        :class:`ansys.dpf.core.scoping.Scoping`

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> my_scoping = elements.scoping

        """
        return self._mesh._get_scoping(loc=locations.elemental)

    @property
    def element_types_field(self):
        """Field of all element types.

        Returns
        -------
        :class:`ansys.dpf.core.field.Field`
            Field of all element types.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> field = elements.element_types_field
        >>> field.data
        array([1, 1, 1, 1, 1, 1, 1, 1])

        """
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.elemental_property = meshed_region_pb2.ElementalPropertyType.ELEMENT_TYPE
        request.elemental_property = meshed_region_pb2.ELEMENT_TYPE
        fieldOut = self._mesh._stub.ListProperty(request)
        return field.Field(server=self._mesh._server, field=fieldOut)

    @property
    @protect_grpc
    def materials_field(self):
        """Field of all material IDs.

        Returns
        -------
        :class:`ansys.dpf.core.field.Field`
            Field of all materials IDs.

        Examples
        --------
        Extract the material IDs from the materials_field

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> elements.materials_field.data
        array([1, 1, 1, 1, 1, 1, 1, 1])

        """
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.elemental_property = meshed_region_pb2.ElementalPropertyType.MATERIAL
        request.elemental_property = meshed_region_pb2.MATERIAL
        fieldOut = self._mesh._stub.ListProperty(request)
        return field.Field(server=self._mesh._server, field=fieldOut)

    @property
    def connectivities_field(self):
        """Field containing for each element ID the node indices connected to the element.

        Returns
        -------
        :class:`ansys.dpf.core.field.Field`
            Field containing for each element ID the node indices connected to the element.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> field = elements.connectivities_field
        >>> field.get_entity_data(1)
        array([ 0, 11, 13, 25,  2,  9,  8,  3, 29, 58, 63, 32, 40, 52, 42, 37, 28,
               55, 53, 43])

        """
        return self._get_connectivities_field()

    @protect_grpc
    def _get_connectivities_field(self):
        """Retrieve the connectivities field."""
        request = meshed_region_pb2.ListPropertyRequest()
        request.mesh.CopyFrom(self._mesh._message)
        # request.elemental_property = meshed_region_pb2.ElementalPropertyType.CONNECTIVITY
        request.elemental_property = meshed_region_pb2.CONNECTIVITY
        fieldOut = self._mesh._stub.ListProperty(request)
        return property_field.PropertyField(
            server=self._mesh._server, property_field=fieldOut
        )

    @property
    def n_elements(self) -> int:
        """Number of elements"""
        return self._mesh._stub.List(self._mesh._message).num_element

    def _build_mapping_id_to_index(self):
        """Retrieve the mapping between the IDs and indices of the entity."""
        return {eid: i for i, eid in enumerate(self.scoping.ids)}

    @property
    def mapping_id_to_index(self) -> dict:
        """Mapping between the IDs and indices of the entity.

        This proprty is useful for mapping scalar results from a field to the meshed region.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.simple_bar)
        >>> meshed_region = model.metadata.meshed_region
        >>> map = meshed_region.nodes.mapping_id_to_index

        """
        if self._mapping_id_to_index is None:
            self._mapping_id_to_index = self._build_mapping_id_to_index()
        return self._mapping_id_to_index

    def map_scoping(self, external_scope):
        """Retrieve the indices to map the scoping of these elements to
        the scoping of a field.

        Parameters
        ----------
        external_scope : :class:`ansys.dpf.core.scoping.Scoping`
            Scoping to map to.

        Returns
        -------
        indices : numpy.ndarray
            List of indices to map from the external scope to the
            scoping of these elements.
        mask : numpy.ndarray
            Members of the external scope that are in the element scoping.

        Examples
        --------
        Return the indices that map a field to an elements collection.

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> vol = model.results.elemental_volume()
        >>> field = vol.outputs.fields_container()[0]
        >>> ind, mask = elements.map_scoping(field.scoping)
        >>> ind
        array([0, 1, 2, 3, 4, 5, 6, 7])

        """
        if external_scope.location in ["Nodal", "NodalElemental"]:
            raise ValueError('Input scope location must be "Nodal"')
        arr = np.array(list(map(self.mapping_id_to_index.get, external_scope.ids)))
        mask = arr != None
        ind = arr[mask].astype(np.int)
        return ind, mask

    @property
    def has_shell_elements(self) -> bool:
        """Whether at least one element is a 2D element (shell).

        Returns
        -------
        bool

        """
        return self._mesh._stub.List(
            self._mesh._message
        ).element_shape_info.has_shell_elements

    @property
    def has_solid_elements(self) -> bool:
        """Whether at list one element is a 3D element (solid).

        Returns
        -------
        bool

        """
        return self._mesh._stub.List(
            self._mesh._message
        ).element_shape_info.has_solid_elements

    @property
    def has_beam_elements(self) -> bool:
        """Whether at least one element is a 1D beam element.

        Returns
        -------
        bool

        """
        return self._mesh._stub.List(
            self._mesh._message
        ).element_shape_info.has_beam_elements

    @property
    def has_point_elements(self) -> bool:
        """Whether at least one element is a point element.

        Returns
        -------
        bool

        """
        return self._mesh._stub.List(
            self._mesh._message
        ).element_shape_info.has_point_elements


class ElementAdder:
    """Provides for adding new elements in a meshed region.

    Parameters
    ----------
    id : int
        ID to assign to the new element.
    shape : str
        Shape of the element. Options are ``"solid"``, ``"shell"``, ``"beam"``
        and ``"unknown_shape"``.
    connectivity : list
        List of the node indices to connect to the new element.

    Examples
    --------
    Create a meshed region from scratch.

    >>> import ansys.dpf.core as dpf
    >>> meshed_region = dpf.MeshedRegion(num_nodes=4, num_elements=1)
    >>> i=0
    >>> for node in meshed_region.nodes.add_nodes(4):
    ...     node.id = i+1
    ...     node.coordinates = [float(i), float(i), 0.0]
    ...     i=i+1
    >>> for element in meshed_region.elements.add_elements(1):
    ...     element.id=1
    ...     element.connectivity = range(0,4)
    ...     element.is_shell=True #or is_solid, is_beam, is_point

    """

    def __init__(self):
        self.id = 0
        self.connectivity = [0]
        self._shape_info = {
            "solid": True,
            "shell": False,
            "beam": False,
            "point": False,
        }

    @property
    def is_solid(self) -> bool:
        """Whether the element is a solid.

        Returns
        -------
        bool
        """

        return self._shape_info["solid"]

    @is_solid.setter
    def is_solid(self, value):
        for key in self._shape_info:
            self._shape_info[key] = False
        self._shape_info["solid"] = True

    @property
    def is_shell(self) -> bool:
        """Whether the element is a shell.

        Returns
        -------
        bool
        """

        return self._shape_info["shell"]

    @is_shell.setter
    def is_shell(self, value):
        for key in self._shape_info:
            self._shape_info[key] = False
        self._shape_info["shell"] = True

    @property
    def is_beam(self) -> bool:
        """Whether the element is a beam.

        Returns
        -------
        bool
        """

        return self._shape_info["beam"]

    @is_beam.setter
    def is_beam(self, value):
        for key in self._shape_info:
            self._shape_info[key] = False
        self._shape_info["beam"] = True

    @property
    def is_point(self) -> bool:
        """Whether the element is a point.

        Returns
        -------
        bool
        """

        return self._shape_info["point"]

    @is_point.setter
    def is_point(self, value):
        for key in self._shape_info:
            self._shape_info[key] = False
        self._shape_info["point"] = True

    @property
    def shape(self) -> str:
        """Shape of the element.

        Returns
        --------
        str
           Shape of the element. Options are ``"solid"``, ``"shell"``, ``"beam"``
           and ``"unknown_shape"``.

        """
        if self.is_solid:
            return "solid"
        elif self.is_shell:
            return "shell"
        elif self.is_beam:
            return "beam"
        else:
            return "unknown_shape"

    @shape.setter
    def shape(self, value):
        """Set the shape of the element.

        Parameters
        --------
        value : str
           Shape of the element. Options are ``"solid"``, ``"shell"``, ``"beam"``
           and ``"unknown_shape"``.

        """
        self._shape_info = {
            "solid": False,
            "shell": False,
            "beam": False,
            "point": False,
        }
        if value == "solid":
            self.is_solid = True
        elif value == "shell":
            self.is_shell = True
        elif value == "beam":
            self.is_beam = True
        else:
            self.is_point = True


class element_types(Enum):
    """Contains the types of elements."""

    General = -2
    All = -1
    Tet10 = 0
    Hex20 = 1
    Wedge15 = 2
    Pyramid13 = 3
    Tri6 = 4
    TriShell6 = 5
    Quad8 = 6
    QuadShell8 = 7
    Line3 = 8
    Point1 = 9
    Tet4 = 10
    Hex8 = 11
    Wedge6 = 12
    Pyramid5 = 13
    Tri3 = 14
    TriShell3 = 15
    Quad4 = 16
    QuadShell4 = 17
    Line2 = 18
    NumElementTypes = 19
    Unknown = 20
    EMagLine = 21
    EMagArc = 22
    EMagCircle = 23
    Surface3 = 24
    Surface4 = 25
    Surface6 = 26
    Surface8 = 27
    Edge2 = 28
    Edge3 = 29
    Beam3 = 30
    Beam4 = 31
    GeneralPlaceholder = 32
    Polygon = 33
    Polyhedron = 34

    @staticmethod
    def _descriptors():
        return {
            element_types.General: ElementDescriptor(
                element_types.General, "General", "general"
            ),
            element_types.All: ElementDescriptor(
                element_types.All, "Unknown", "unknown"
            ),
            element_types.Tet10: ElementDescriptor(
                element_types.Tet10,
                "Quadratic 10-nodes Tetrahedron",
                "tet10",
                "solid",
                4,
                6,
                10,
                True,
                False,
                False,
                True,
            ),
            element_types.Hex20: ElementDescriptor(
                element_types.Hex20,
                "Quadratic 20-nodes Hexa",
                "hex20",
                "solid",
                8,
                12,
                20,
                True,
                False,
                False,
                True,
            ),
            element_types.Wedge15: ElementDescriptor(
                element_types.Wedge15,
                "Quadratic 15-nodes Wedge",
                "wedge15",
                "solid",
                6,
                9,
                15,
                True,
                False,
                False,
                True,
            ),
            element_types.Pyramid13: ElementDescriptor(
                element_types.Pyramid13,
                "Quadratic 13-nodes Pyramid",
                "pyramid13",
                "solid",
                5,
                8,
                13,
                True,
                False,
                False,
                True,
            ),
            element_types.Tri6: ElementDescriptor(
                element_types.Tri6,
                "Quadratic 6-nodes Triangle",
                "tri6",
                "shell",
                3,
                3,
                6,
                False,
                True,
                False,
                True,
            ),
            element_types.TriShell6: ElementDescriptor(
                element_types.TriShell6,
                "Quadratic 6-nodes Triangle Shell",
                "triShell6",
                "shell",
                3,
                3,
                6,
                False,
                True,
                False,
                True,
            ),
            element_types.Quad8: ElementDescriptor(
                element_types.Quad8,
                "Quadratic 8-nodes Quadrangle",
                "quad8",
                "shell",
                4,
                4,
                8,
                False,
                True,
                False,
                True,
            ),
            element_types.QuadShell8: ElementDescriptor(
                element_types.QuadShell8,
                "Quadratic 8-nodes Quadrangle Shell",
                "quadShell8",
                "shell",
                4,
                4,
                8,
                False,
                True,
                False,
                True,
            ),
            element_types.Line3: ElementDescriptor(
                element_types.Line3,
                "Quadratic 3-nodes Line",
                "line3",
                "beam",
                2,
                1,
                3,
                False,
                False,
                True,
                True,
            ),
            element_types.Point1: ElementDescriptor(
                element_types.Point1,
                "Point",
                "point1",
                "point",
                1,
                0,
                1,
                False,
                False,
                False,
                False,
            ),
            element_types.Tet4: ElementDescriptor(
                element_types.Tet4,
                "Linear 4-nodes Tetrahedron",
                "tet4",
                "solid",
                4,
                0,
                4,
                True,
                False,
                False,
                False,
            ),
            element_types.Hex8: ElementDescriptor(
                element_types.Hex8,
                "Linear 8-nodes Hexa",
                "hex8",
                "solid",
                8,
                0,
                8,
                True,
                False,
                False,
                False,
            ),
            element_types.Wedge6: ElementDescriptor(
                element_types.Wedge6,
                "Linear 6-nodes Wedge",
                "wedge6",
                "solid",
                6,
                0,
                6,
                True,
                False,
                False,
                False,
            ),
            element_types.Pyramid5: ElementDescriptor(
                element_types.Pyramid5,
                "Linear 5-nodes Pyramid",
                "pyramid5",
                "solid",
                5,
                0,
                5,
                True,
                False,
                False,
                False,
            ),
            element_types.Tri3: ElementDescriptor(
                element_types.Tri3,
                "Linear 3-nodes Triangle",
                "tri3",
                "shell",
                3,
                0,
                3,
                False,
                True,
                False,
                False,
            ),
            element_types.TriShell3: ElementDescriptor(
                element_types.TriShell3,
                "Linear 3-nodes Triangle Shell",
                "triShell3",
                "shell",
                3,
                0,
                3,
                False,
                True,
                False,
                False,
            ),
            element_types.Quad4: ElementDescriptor(
                element_types.Quad4,
                "Linear 4-nodes Quadrangle",
                "quad4",
                "shell",
                4,
                0,
                4,
                False,
                True,
                False,
                False,
            ),
            element_types.QuadShell4: ElementDescriptor(
                element_types.QuadShell4,
                "Linear 4-nodes Quadrangle Shell",
                "quadShell4",
                "shell",
                4,
                0,
                4,
                False,
                True,
                False,
                False,
            ),
            element_types.Line2: ElementDescriptor(
                element_types.Line2,
                "Linear 2-nodes Line",
                "line2",
                "beam",
                2,
                0,
                2,
                False,
                False,
                True,
                False,
            ),
            element_types.NumElementTypes: ElementDescriptor(
                element_types.NumElementTypes, "NumElementTypes", "numElementTypes"
            ),
            element_types.Unknown: ElementDescriptor(
                element_types.Unknown, "Unknown", "unknown"
            ),
            element_types.EMagLine: ElementDescriptor(
                element_types.EMagLine, "EMagLine", "EMagLine", "beam"
            ),
            element_types.EMagArc: ElementDescriptor(
                element_types.EMagArc, "EMagArc", "EMagArc", "beam"
            ),
            element_types.EMagCircle: ElementDescriptor(
                element_types.EMagCircle, "EMagCircle", "EMagCircle", "shell"
            ),
            element_types.Surface3: ElementDescriptor(
                element_types.Surface3, "Surface3", "surface3", "shell"
            ),
            element_types.Surface4: ElementDescriptor(
                element_types.Surface4, "Surface4", "surface4", "shell"
            ),
            element_types.Surface6: ElementDescriptor(
                element_types.Surface6, "Surface6", "surface6", "shell"
            ),
            element_types.Surface8: ElementDescriptor(
                element_types.Surface8, "Surface8", "surface8", "shell"
            ),
            element_types.Edge2: ElementDescriptor(
                element_types.Edge2, "Edge2", "edge2", "beam"
            ),
            element_types.Edge3: ElementDescriptor(
                element_types.Edge3, "Edge3", "edge3", "beam"
            ),
            element_types.Beam3: ElementDescriptor(
                element_types.Beam3, "Beam3", "beam3", "beam"
            ),
            element_types.Beam4: ElementDescriptor(
                element_types.Beam4, "Beam4", "beam4", "beam"
            ),
            element_types.GeneralPlaceholder: ElementDescriptor(
                element_types.GeneralPlaceholder,
                "GeneralPlaceholder",
                "generalPlaceholder",
            ),
            element_types.Polygon: ElementDescriptor(
                element_types.Polygon,
                "Polygon",
                "polygon",
                "shell",
                -1,
                0,
                -1,
                False,
                True,
                False,
            ),
            element_types.Polyhedron: ElementDescriptor(
                element_types.Polyhedron,
                "Polyhedron",
                "polyhedron",
                "solid",
                -1,
                0,
                -1,
                True,
                False,
                False,
            ),
        }

    @staticmethod
    def shape(element_type):
        """Retrieve the shape of the element.

        Returns
        -------
        type

        """
        if isinstance(element_type, (int, np.int32)):
            element_type = element_types(element_type)
        el_shape = element_types._descriptors().get(element_type, None).shape
        return el_shape

    @staticmethod
    def descriptor(element_type):
        """Retrieve element information.

        This method provides access to an instance of the ``ElementDescriptor`` of the requested
        element to retrieve such information as the number of nodes and shape.

        Parameters
        ----------
        element_type : int
            Type of the elemment.

        Returns
        -------
        ElementDescriptor
            Element descriptor instance that provides element information.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> elements = model.metadata.meshed_region.elements
        >>> element = elements[0]
        >>> type_of_element = element.type
        >>> type_of_element
        <element_types.Hex20: 1>
        >>> element_descriptor = dpf.element_types.descriptor(type_of_element)
        >>> element_descriptor.name
        'hex20'
        >>> element_descriptor.description
        'Quadratic 20-nodes Hexa'

        """
        if isinstance(element_type, (int, np.int32)):
            element_type = element_types(element_type)
        descriptor = element_types._descriptors().get(element_type, None)
        return descriptor


element_types.__doc__ = __write_enum_doc__(
    element_types, "Types of elements available in a dpf's mesh."
)



"""
forward_meshes_container
========================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "utility" category
"""

class forward_meshes_container(Operator):
    """Return the input mesh or meshes container into a meshes container.

      available inputs:
        - meshes (MeshesContainer, MeshedRegion)
        - default_label (str) (optional)

      available outputs:
        - meshes_container (MeshesContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.utility.forward_meshes_container()

      >>> # Make input connections
      >>> my_meshes = dpf.MeshesContainer()
      >>> op.inputs.meshes.connect(my_meshes)
      >>> my_default_label = str()
      >>> op.inputs.default_label.connect(my_default_label)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.utility.forward_meshes_container(meshes=my_meshes,default_label=my_default_label)

      >>> # Get output data
      >>> result_meshes_container = op.outputs.meshes_container()"""
    def __init__(self, meshes=None, default_label=None, config=None, server=None):
        super().__init__(name="forward_meshes_container", config = config, server = server)
        self._inputs = InputsForwardMeshesContainer(self)
        self._outputs = OutputsForwardMeshesContainer(self)
        if meshes !=None:
            self.inputs.meshes.connect(meshes)
        if default_label !=None:
            self.inputs.default_label.connect(default_label)

    @staticmethod
    def _spec():
        spec = Specification(description="""Return the input mesh or meshes container into a meshes container.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "meshes", type_names=["meshes_container","abstract_meshed_region"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "default_label", type_names=["string"], optional=True, document="""this default label is used if a new meshes container needs to be created (default is unknown)""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "meshes_container", type_names=["meshes_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "forward_meshes_container")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsForwardMeshesContainer 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsForwardMeshesContainer 
        """
        return super().outputs


#internal name: forward_meshes_container
#scripting name: forward_meshes_container
class InputsForwardMeshesContainer(_Inputs):
    """Intermediate class used to connect user inputs to forward_meshes_container operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.utility.forward_meshes_container()
      >>> my_meshes = dpf.MeshesContainer()
      >>> op.inputs.meshes.connect(my_meshes)
      >>> my_default_label = str()
      >>> op.inputs.default_label.connect(my_default_label)
    """
    def __init__(self, op: Operator):
        super().__init__(forward_meshes_container._spec().inputs, op)
        self._meshes = Input(forward_meshes_container._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._meshes)
        self._default_label = Input(forward_meshes_container._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._default_label)

    @property
    def meshes(self):
        """Allows to connect meshes input to the operator

        Parameters
        ----------
        my_meshes : MeshesContainer, MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.utility.forward_meshes_container()
        >>> op.inputs.meshes.connect(my_meshes)
        >>> #or
        >>> op.inputs.meshes(my_meshes)

        """
        return self._meshes

    @property
    def default_label(self):
        """Allows to connect default_label input to the operator

        - pindoc: this default label is used if a new meshes container needs to be created (default is unknown)

        Parameters
        ----------
        my_default_label : str, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.utility.forward_meshes_container()
        >>> op.inputs.default_label.connect(my_default_label)
        >>> #or
        >>> op.inputs.default_label(my_default_label)

        """
        return self._default_label

class OutputsForwardMeshesContainer(_Outputs):
    """Intermediate class used to get outputs from forward_meshes_container operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.utility.forward_meshes_container()
      >>> # Connect inputs : op.inputs. ...
      >>> result_meshes_container = op.outputs.meshes_container()
    """
    def __init__(self, op: Operator):
        super().__init__(forward_meshes_container._spec().outputs, op)
        self._meshes_container = Output(forward_meshes_container._spec().output_pin(0), 0, op) 
        self._outputs.append(self._meshes_container)

    @property
    def meshes_container(self):
        """Allows to get meshes_container output of the operator


        Returns
        ----------
        my_meshes_container : MeshesContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.utility.forward_meshes_container()
        >>> # Connect inputs : op.inputs. ...
        >>> result_meshes_container = op.outputs.meshes_container() 
        """
        return self._meshes_container




"""
Workflow
========
Interface to underlying gRPC workflow.
"""
import logging

from ansys import dpf
from ansys.dpf.core import dpf_operator, inputs, outputs
from ansys.dpf.core.errors import protect_grpc
from ansys.grpc.dpf import base_pb2, workflow_pb2, workflow_pb2_grpc

LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")


class Workflow:
    """Represents a workflow.

    A workflow is a black box containing operators and exposing only the necessary operator's
    inputs and outputs to compute a given algorithm.

    Parameters
    ----------
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
    workflow :  workflow_pb2.Workflow

    Examples
    --------
    Create a generic workflow computing the minimum of displacement by chaining the ``'U'``
    and ``'min_max_fc'`` operators.

    >>> from ansys.dpf import core as dpf
    >>> disp_op = dpf.operators.result.displacement()
    >>> max_fc_op = dpf.operators.min_max.min_max_fc(disp_op)
    >>> workflow = dpf.Workflow()
    >>> workflow.add_operators([disp_op,max_fc_op])
    >>> workflow.set_input_name("data_sources", disp_op.inputs.data_sources)
    >>> workflow.set_output_name("min", max_fc_op.outputs.field_min)
    >>> workflow.set_output_name("max", max_fc_op.outputs.field_max)


    >>> from ansys.dpf.core import examples
    >>> data_src = dpf.DataSources(examples.multishells_rst)
    >>> workflow.connect("data_sources", data_src)
    >>> min = workflow.get_output("min", dpf.types.field)
    >>> max = workflow.get_output("max", dpf.types.field)

    """

    def __init__(self, workflow=None, server=None):
        """Initialize the workflow by connecting to a stub."""
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        self._message = workflow

        if workflow is None:
            self.__send_init_request()

    @protect_grpc
    def connect(self, pin_name, inpt, pin_out=0):
        """Connect an input on the workflow using a pin name.

        Parameters
        ----------
        pin_name : str
            Name of the pin to connect. This name should be
            exposed before with wf.set_input_name
        inpt : str, int, double, bool, list of int, list of doubles,
               Field, FieldsContainer, Scoping, ScopingsContainer,
        MeshedRegion, MeshesContainer, DataSources, Operator
            Object to connect to.
        pin_out : int, optional
            If the input is an operator, the output pin of the input operator.
            The default is ``0``.

        Examples
        --------
        Create a generic workflow computing the minimum of displacement by chaining the ``'U'``
        and ``'min_max_fc'`` operators.

        >>> from ansys.dpf import core as dpf
        >>> disp_op = dpf.operators.result.displacement()
        >>> max_fc_op = dpf.operators.min_max.min_max_fc(disp_op)
        >>> workflow = dpf.Workflow()
        >>> workflow.add_operators([disp_op,max_fc_op])
        >>> workflow.set_input_name("data_sources", disp_op.inputs.data_sources)
        >>> workflow.set_output_name("min", max_fc_op.outputs.field_min)
        >>> workflow.set_output_name("max", max_fc_op.outputs.field_max)


        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> workflow.connect("data_sources", data_src)
        >>> min = workflow.get_output("min", dpf.types.field)
        >>> max = workflow.get_output("max", dpf.types.field)

        """
        request = workflow_pb2.UpdateConnectionRequest()
        request.wf.CopyFrom(self._message)
        request.pin_name = pin_name
        dpf_operator._fillConnectionRequestMessage(request, inpt, pin_out)
        self._stub.UpdateConnection(request)

    @protect_grpc
    def get_output(self, pin_name, output_type):
        """Retrieve the output of the operator on the pin number.

        Parameters
        ----------
        pin_name : str
            Name of the pin to retrieve. This name should be
            exposed before with wf.set_output_name
        output_type : core.type enum
            Type of the requested output.
        """

        request = workflow_pb2.WorkflowEvaluationRequest()
        request.wf.CopyFrom(self._message)
        request.pin_name = pin_name

        if output_type is not None:
            dpf_operator._write_output_type_to_proto_style(output_type, request)
            out = self._stub.Get(request)
            return dpf_operator._convertOutputMessageToPythonInstance(
                out, output_type, self._server
            )
        else:
            raise ValueError(
                "please specify an output type to get the workflow's output"
            )

    def set_input_name(self, name, *args):
        """Set the name of the input pin of the workflow to expose it for future connection.

        Parameters
        ----------
        name : str
            Name of the pin to connect. This name should be
            exposed before with wf.set_input_name
        *args : core.Operator, core.Input, int
            Operator with its input pin number or input to name.

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.operators.result.displacement()
        >>> max_fc_op = dpf.operators.min_max.min_max_fc(disp_op)
        >>> workflow.set_input_name("data_sources", disp_op.inputs.data_sources)

        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> workflow.connect("data_sources", data_src)

        """
        request = workflow_pb2.UpdatePinNamesRequest()
        request.wf.CopyFrom(self._message)
        input_request = workflow_pb2.OperatorNaming()
        input_request.name = name
        input_request.pin = 0
        for arg in args:
            if isinstance(arg, inputs.Input):
                input_request.pin = arg._pin
                input_request.operator.CopyFrom(arg._operator._message)
            elif isinstance(arg, dpf_operator.Operator):
                input_request.operator.CopyFrom(arg._message)
            elif isinstance(arg, int):
                input_request.pin = arg
        request.inputs_naming.extend([input_request])
        self._stub.UpdatePinNames(request)

    def set_output_name(self, name, *args):
        """Set the name of the output pin of the workflow to expose it for future connection.

        Parameters
        ----------
        name : str
            Name of the pin to connect. This name should be
            exposed before with wf.set_input_name
        *args : core.Operator, core.Output, int
            Operator with its outpt pin number or output to name.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> workflow = dpf.Workflow()
        >>> model = dpf.Model(examples.simple_bar)
        >>> disp_op = model.results.displacement()
        >>> max_fc_op = dpf.operators.min_max.min_max_fc(disp_op)
        >>> workflow.set_output_name("contour", disp_op.outputs.fields_container)
        >>> fc = workflow.get_output("contour", dpf.types.fields_container)

        """
        request = workflow_pb2.UpdatePinNamesRequest()
        request.wf.CopyFrom(self._message)
        output_request = workflow_pb2.OperatorNaming()
        output_request.name = name
        output_request.pin = 0
        for arg in args:
            if isinstance(arg, outputs.Output):
                output_request.pin = arg._pin
                output_request.operator.CopyFrom(arg._operator._message)
            elif isinstance(arg, dpf_operator.Operator):
                output_request.operator.CopyFrom(arg._message)
            elif isinstance(arg, int):
                output_request.pin = arg
        request.outputs_naming.extend([output_request])
        self._stub.UpdatePinNames(request)

    def add_operators(self, operators):
        """Add operators to the list of operators of the workflow.

        Parameters
        ----------
        operators : dpf.core.Operator, list of dpf.core.Operator
            Operators to add to the list.

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> max_op = dpf.Operator("min_max")
        >>> workflow.add_operator([disp_op,max_op])

        """
        request = workflow_pb2.AddOperatorsRequest()
        request.wf.CopyFrom(self._message)
        if isinstance(operators, list):
            request.operators.extend([op._message for op in operators])
        elif isinstance(operators, dpf_operator.Operator):
            request.operators.extend([operators._message])
        else:
            raise TypeError(
                "Operators to add to the workflow are expected to be of "
                f"type {type(list).__name__} or {type(dpf_operator.Operator).__name__}"
            )
        self._stub.AddOperators(request)

    def add_operator(self, operator):
        """Add an operator to the list of operators of the workflow.

        Parameters
        ----------
        operator : dpf.core.Operator
            Operator to add to the list.

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> workflow.add_operator(disp_op)

        """
        self.add_operators(operator)

    def record(self, identifier=None, transfer_ownership=True):
        """Add the workflow to DPF's internal registry with an ID returned by this method.

        The workflow can be recovered by ``dpf.core.Workflow.get_recorded_workflow(id)``.

        Parameters
        ----------
        identifier : str, optional
            Name given to the workflow.
        transfer_ownership : bool
            Whether to transfer the ownership. The default is ``True``. If the ownership is
            not transferred, the workflow is removed from the internal registry
            as soon as the workflow has been recovered by its ID.

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> workflow.add_operator(disp_op)
        >>> # ...
        >>> id = workflow.record()
        >>> workflow_copy = dpf.Workflow.get_recorded_workflow(id)

        """
        request = workflow_pb2.RecordInInternalRegistryRequest()
        request.wf.CopyFrom(self._message)
        if identifier:
            request.identifier = identifier
        request.transferOwnership = transfer_ownership
        return self._stub.RecordInInternalRegistry(request).id

    @staticmethod
    def get_recorded_workflow(id, server=None):
        """Retrieve a workflow registered (with workflow.record())

        Parameters
        ----------
        id : int
            ID given by the method "record".

        Returns
        ----------
        workflow : core.Workflow()
            workflow registered in dpf's registry (server side)

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> workflow = dpf.Workflow()
        >>> disp_op = dpf.Operator("U")
        >>> workflow.add_operator(disp_op)
        >>> # ...
        >>> id = workflow.record()
        >>> workflow_copy = dpf.Workflow.get_recorded_workflow(id)

        """
        request = workflow_pb2.WorkflowFromInternalRegistryRequest()
        request.registry_id = id
        wf = Workflow(server=server)
        wf._message.CopyFrom(wf._stub.GetFromInternalRegistry(request))
        return wf

    @property
    def info(self):
        """Dictionary with the operator names and the exposed input and output names.

        Returns
        ----------
        info : dictionarry str->list str
            Dictionary with ``"operator_names"``, ``"input_names"``, and ``"output_names"`` key.
        """
        tmp = self._stub.List(self._message)
        out = {"operator_names": [], "input_names": [], "output_names": []}
        for name in tmp.operator_names:
            out["operator_names"].append(name)
        for name in tmp.input_pin_names.pin_names:
            out["input_names"].append(name)
        for name in tmp.output_pin_names.pin_names:
            out["output_names"].append(name)
        return out

    @property
    def operator_names(self):
        """List of the names of operators added in the workflow.

        Returns
        ----------
        names : list str
        """
        return self.info["operator_names"]

    @property
    def input_names(self):
        """List of the input names exposed in the workflow with set_input_name.

        Returns
        ----------
        names : list str
        """
        return self.info["input_names"]

    @property
    def output_names(self):
        """List of the output names exposed in the workflow with set_output_name.

        Returns
        ----------
        names : list str
        """
        return self.info["output_names"]

    def chain_with(self, workflow, input_output_names=None):
        """Chain two workflows together so that they become one workflow.

        The one workflow contains all the operators, inputs, and outputs
        exposed in both workflows.

        Parameters
        ----------
        workflow : core.Workflow
            Second workflow's inputs to chained with this workflow's outputs.
        input_output_names : str tuple, optional
            Input name of the workflow to chain with the output name of the second workflow.
            The default is ``None``, in which case this outputs in this workflow with the same
            names as the inputs in the second workflow are chained.

        Examples
        --------
        ::

            +-------------------------------------------------------------------------------------------------+
            |  INPUT:                                                                                         |
            |                                                                                                 |
            |input_output_names = ("output","field" )                                                         |
            |                      ____                                  ______________________               |
            |  "data_sources"  -> |this| ->  "stuff"        "field" -> |workflow_to_chain_with| -> "contour"  |
            |"time_scoping"    -> |    |             "mesh_scoping" -> |                      |               |
            |                     |____| ->  "output"                  |______________________|               |
            |  OUTPUT                                                                                         |
            |                    ____                                                                         |
            |"data_sources"  -> |this| ->  "stuff"                                                            |
            |"time_scoping" ->  |    | ->  "contour"                                                          |
            |"mesh_scoping" ->  |____| -> "output"                                                            |
            +-------------------------------------------------------------------------------------------------+ # noqa: E501


        """
        request = workflow_pb2.ChainRequest()
        request.wf.CopyFrom(self._message)
        request.wf_to_chain_with.CopyFrom(workflow._message)
        if input_output_names:
            request.input_to_output.output_name = input_output_names[0]
            request.input_to_output.input_name = input_output_names[1]
        self._stub.Chain(request)

    def _connect(self):
        """Connect to the gRPC service."""
        return workflow_pb2_grpc.WorkflowServiceStub(self._server.channel)

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description

        return _description(self._message, self._server)

    @protect_grpc
    def __send_init_request(self):
        request = base_pb2.Empty()
        self._message = self._stub.Create(request)



"""
time_freq_scoping_factory
=========================

Contains functions to simplify creating time frequency scopings.
"""

from ansys.dpf.core import Scoping
from ansys.dpf.core import errors as dpf_errors
from ansys.dpf.core.common import locations


def scoping_by_load_step(load_step, server=None):
    """Create a specific ``ansys.dpf.core.Scoping`` for a given load step.

    The returned scoping describes a specific time frequencey support element
    for a given load step.

    Parameters
    ----------
    load_step : int
        Load step ID of the specific time frequency scoiping.
     server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting one load step.
    """
    scoping = Scoping(server=server, ids=[load_step], location=locations.time_freq_step)
    return scoping


def scoping_by_load_steps(load_steps, server=None):
    """Create a specific :class:`ansys.dpf.core.Scoping` for a given list of load steps.

    The returned scoping describes a specific time frequency support element
    for a given list of load steps.

    Parameters
    ----------
    load_steps : list of int
        List of load steps IDs of the specific time frequency scoping.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting several load_steps.
    """
    if not isinstance(load_steps, list):
        raise dpf_errors.InvalidTypeError("list", "load_steps")
    scoping = Scoping(server=server, ids=load_steps, location=locations.time_freq_step)
    return scoping


def scoping_by_set(cumulative_set, server=None):
    """Create a specific :class:`ansys.dpf.core.Scoping` for a given cumulative set index.

    The returned scoping describes a specific time frequency support element for a given
    cumulative set index.

    Parameters
    ----------
    cumulative_set : int
        Cumulative index of the set.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting one set (referenced by cumulative index).
    """
    scoping = Scoping(server=server, ids=[cumulative_set], location=locations.time_freq)
    return scoping


def scoping_by_sets(cumulative_sets, server=None):
    """Create a specific :class:`ansys.dpf.core.Scoping` for a given list of cumulative set indices.

    The returned scoping describes a specific time frequencey support element for a given
    list of cumulative sets of indices.

    Parameters
    ----------
    cumulative_sets : list of int
        List of cumulative indices of the sets.
     server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping targeting severals sets (referenced by cumulative indices).
    """
    if not isinstance(cumulative_sets, list):
        raise dpf_errors.InvalidTypeError("list", "cumulative_sets")
    scoping = Scoping(server=server, ids=cumulative_sets, location=locations.time_freq)
    return scoping


def scoping_by_step_and_substep(
    load_step_id, subset_id, time_freq_support, server=None
):
    """Create a specific :class:`ansys.dpf.core.Scoping` for a given step and subset.

    The returned scoping describes a specific time frequency support element for a given
    step and substep.

    Parameters
    ----------
    load_step_id : int
        ID of the load step.
    subset_id : int
        ID of the subset.
    time_freq_support : ansys.dpf.core.TimeFreqSupport
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping based on a given step and substep of a time frequency support.
    """
    set_index = time_freq_support.get_cumulative_index(load_step_id - 1, subset_id - 1)
    scoping = Scoping(ids=[set_index + 1], location=locations.time_freq)
    return scoping


def scoping_by_step_and_substep_from_model(load_step_id, subset_id, model, server=None):
    """Create a specific ``ansys.dpf.core.Scoping`` for a given step and substep.

    The returned scoping describes a specific model's time freq support element for a given
    step and substep.

    Parameters
    ----------
    load_step_id : int
        ID of the step.
    subset_id : int
        ID of the subset.
    time_freq_support : ansys.dpf.core.TimeFreqSupport()
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    scoping : ansys.dpf.core.Scoping
        Scoping based on a given step/substep of a model's time_freq_support."""
    return scoping_by_step_and_substep(
        load_step_id, subset_id, model.metadata.time_freq_support
    )



"""
node_coordinates
================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "mesh" category
"""

class node_coordinates(Operator):
    """Returns the node coordinates of the mesh(es) in input

      available inputs:
        - mesh (MeshedRegion, MeshesContainer)

      available outputs:
        - coordinates (Field ,FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.mesh.node_coordinates()

      >>> # Make input connections
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.mesh.node_coordinates(mesh=my_mesh)

      >>> # Get output data
      >>> result_coordinates = op.outputs.coordinates()"""
    def __init__(self, mesh=None, config=None, server=None):
        super().__init__(name="mesh::node_coordinates", config = config, server = server)
        self._inputs = InputsNodeCoordinates(self)
        self._outputs = OutputsNodeCoordinates(self)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        spec = Specification(description="""Returns the node coordinates of the mesh(es) in input""",
                             map_input_pin_spec={
                                 7 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region","meshes_container"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "coordinates", type_names=["field","fields_container"], optional=False, document="""if the input is a meshed region, a field of coordinates is the output, else if the input is a  meshes container, a fields container (one field by mesh) is the output""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "mesh::node_coordinates")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsNodeCoordinates 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsNodeCoordinates 
        """
        return super().outputs


#internal name: mesh::node_coordinates
#scripting name: node_coordinates
class InputsNodeCoordinates(_Inputs):
    """Intermediate class used to connect user inputs to node_coordinates operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.node_coordinates()
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
    """
    def __init__(self, op: Operator):
        super().__init__(node_coordinates._spec().inputs, op)
        self._mesh = Input(node_coordinates._spec().input_pin(7), 7, op, -1) 
        self._inputs.append(self._mesh)

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        Parameters
        ----------
        my_mesh : MeshedRegion, MeshesContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mesh.node_coordinates()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

class OutputsNodeCoordinates(_Outputs):
    """Intermediate class used to get outputs from node_coordinates operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mesh.node_coordinates()
      >>> # Connect inputs : op.inputs. ...
      >>> result_coordinates = op.outputs.coordinates()
    """
    def __init__(self, op: Operator):
        super().__init__(node_coordinates._spec().outputs, op)
        self.coordinates_as_field = Output( _modify_output_spec_with_one_type(node_coordinates._spec().output_pin(0), "field"), 0, op) 
        self._outputs.append(self.coordinates_as_field)
        self.coordinates_as_fields_container = Output( _modify_output_spec_with_one_type(node_coordinates._spec().output_pin(0), "fields_container"), 0, op) 
        self._outputs.append(self.coordinates_as_fields_container)




"""
component_selector
==================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "logic" category
"""

class component_selector(Operator):
    """Create a scalar/vector field based on the selected component.

      available inputs:
        - field (Field, FieldsContainer)
        - component_number (int, list)
        - default_value (float) (optional)

      available outputs:
        - field (Field)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.logic.component_selector()

      >>> # Make input connections
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_component_number = int()
      >>> op.inputs.component_number.connect(my_component_number)
      >>> my_default_value = float()
      >>> op.inputs.default_value.connect(my_default_value)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.logic.component_selector(field=my_field,component_number=my_component_number,default_value=my_default_value)

      >>> # Get output data
      >>> result_field = op.outputs.field()"""
    def __init__(self, field=None, component_number=None, default_value=None, config=None, server=None):
        super().__init__(name="component_selector", config = config, server = server)
        self._inputs = InputsComponentSelector(self)
        self._outputs = OutputsComponentSelector(self)
        if field !=None:
            self.inputs.field.connect(field)
        if component_number !=None:
            self.inputs.component_number.connect(component_number)
        if default_value !=None:
            self.inputs.default_value.connect(default_value)

    @staticmethod
    def _spec():
        spec = Specification(description="""Create a scalar/vector field based on the selected component.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field","fields_container"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "component_number", type_names=["int32","vector<int32>"], optional=False, document="""one or several component index that will be extracted from the initial field."""), 
                                 2 : PinSpecification(name = "default_value", type_names=["double"], optional=True, document="""set a default value for components that do not exist""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "component_selector")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsComponentSelector 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsComponentSelector 
        """
        return super().outputs


#internal name: component_selector
#scripting name: component_selector
class InputsComponentSelector(_Inputs):
    """Intermediate class used to connect user inputs to component_selector operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.logic.component_selector()
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_component_number = int()
      >>> op.inputs.component_number.connect(my_component_number)
      >>> my_default_value = float()
      >>> op.inputs.default_value.connect(my_default_value)
    """
    def __init__(self, op: Operator):
        super().__init__(component_selector._spec().inputs, op)
        self._field = Input(component_selector._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._field)
        self._component_number = Input(component_selector._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._component_number)
        self._default_value = Input(component_selector._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._default_value)

    @property
    def field(self):
        """Allows to connect field input to the operator

        Parameters
        ----------
        my_field : Field, FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.logic.component_selector()
        >>> op.inputs.field.connect(my_field)
        >>> #or
        >>> op.inputs.field(my_field)

        """
        return self._field

    @property
    def component_number(self):
        """Allows to connect component_number input to the operator

        - pindoc: one or several component index that will be extracted from the initial field.

        Parameters
        ----------
        my_component_number : int, list, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.logic.component_selector()
        >>> op.inputs.component_number.connect(my_component_number)
        >>> #or
        >>> op.inputs.component_number(my_component_number)

        """
        return self._component_number

    @property
    def default_value(self):
        """Allows to connect default_value input to the operator

        - pindoc: set a default value for components that do not exist

        Parameters
        ----------
        my_default_value : float, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.logic.component_selector()
        >>> op.inputs.default_value.connect(my_default_value)
        >>> #or
        >>> op.inputs.default_value(my_default_value)

        """
        return self._default_value

class OutputsComponentSelector(_Outputs):
    """Intermediate class used to get outputs from component_selector operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.logic.component_selector()
      >>> # Connect inputs : op.inputs. ...
      >>> result_field = op.outputs.field()
    """
    def __init__(self, op: Operator):
        super().__init__(component_selector._spec().outputs, op)
        self._field = Output(component_selector._spec().output_pin(0), 0, op) 
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator


        Returns
        ----------
        my_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.logic.component_selector()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field() 
        """
        return self._field




"""
.. _ref_data_sources:

Data Sources
============
"""
import os

from ansys import dpf
from ansys.grpc.dpf import data_sources_pb2, data_sources_pb2_grpc, base_pb2
from ansys.dpf.core.errors import protect_grpc


class DataSources:
    """Contains files with analysis results.

    An extension key (``'rst'`` for example) is used to choose which files represent
    results files versus accessory files. You can set a result file path when
    initializing this class.


    Parameters
    ----------
    result_path : str, optional
        Path of the result. The default is ``None``.
    data_sources : ansys.grpc.dpf.data_sources_pb2.DataSources
        gRPC data sources message. The default is ``None``.
    server : server.DPFServer, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    Initialize a model from a result path.

    >>> from ansys.dpf import core as dpf
    >>> my_data_sources = dpf.DataSources('file.rst')
    >>> my_data_sources.result_files
    ['file.rst']

    """

    def __init__(self, result_path=None, data_sources=None, server=None):
        """Initialize a connection with the server."""
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        if data_sources is None:
            request = base_pb2.Empty()
            self._message = self._stub.Create(request)
        else:
            self._message = data_sources

        if result_path is not None:
            self.set_result_file_path(result_path)

    @protect_grpc
    def _connect(self):
        """Connect to the gRPC service."""
        return data_sources_pb2_grpc.DataSourcesServiceStub(self._server.channel)

    def set_result_file_path(self, filepath, key=""):
        """Add a result file path to the data sources.

        Parameters
        ----------
        filepath : str
            Path to the result file.
        key : str, optional
            Extension of the file, which is used as a key for choosing the correct
            plugin when a result is requested by an operator. The default is ``""``,
            in which case the key is found directly.

        Examples
        --------
        Create a data source and set the result file path.

        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.set_result_file_path('/tmp/file.rst')
        >>> data_sources.result_files
        ['/tmp/file.rst']

        """
        request = data_sources_pb2.UpdateRequest()
        request.result_path = True
        request.key = key
        request.path = filepath
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)

    def set_domain_result_file_path(self, path, domain_id):
        """Add a result file path by domain.

        This method is used to handle files created by a
        distributed solve.

        Parameters
        ----------
        path: str
            Path to the file.
        domain_id: int, optional
            Domain ID for the distributed files.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.set_domain_result_file_path('/tmp/file0.sub', 0)
        >>> data_sources.set_domain_result_file_path('/tmp/file1.sub', 1)

        """
        request = data_sources_pb2.UpdateRequest()
        request.result_path = True
        request.domain.domain_path = True
        request.domain.domain_id = domain_id
        request.path = path
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)

    def add_file_path(self, filepath, key="", is_domain: bool = False, domain_id=0):
        """Add a file path to the data sources.

        Files not added as result files are accessory files, which contain accessory
        information not present in the result files.

        Parameters
        ----------
        filepath : str
            Path of the file.
        key : str, optional
            Extension of the file, which is used as a key for choosing the correct
            plugin when a result is requested by an operator. The default is ``""``,
            in which case the key is found directly.
        is_domain: bool, optional
            Whether the file path is the domain path. The default is ``False``.
        domain_id: int, optional
            Domain ID for the distributed files. The default is ``0``. For this
            parameter to be taken into account, ``domain_path=True`` must be set.
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.add_file_path('/tmp/ds.dat')

        """
        # The filename needs to be a fully qualified file name
        if not os.path.dirname(filepath):
            # append local path
            filepath = os.path.join(os.getcwd(), os.path.basename(filepath))

        request = data_sources_pb2.UpdateRequest()
        request.key = key
        request.path = filepath
        if is_domain:
            request.domain.domain_path = True
            request.domain.domain_id = domain_id
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)

    def add_file_path_for_specified_result(self, filepath, key="", result_key=""):
        """Add a file path for a specified result file key to the data sources.

        This method can be used when results files with different keys (extensions) are
        contained in the data sources. For example, a solve using two different solvers
        could generate two different sets of files.

        Parameters
        ----------
        filepath : str
            Path of the file.
        key : str, optional
            Extension of the file, which is used as a key for choosing the correct
            plugin when a result is requested by an operator. The default is ``""``,
            in which case the key is found directly.
        result_key: str, optional
            Extension of the results file that the specified file path belongs to.
            The default is ``""``, in which case the key is found directly.
        """
        # The filename needs to be a fully qualified file name
        if not os.path.dirname(filepath):
            # append local path
            filepath = os.path.join(os.getcwd(), os.path.basename(filepath))

        request = data_sources_pb2.UpdateRequest()
        request.key = key
        request.result_key = result_key
        request.path = filepath
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)

    def add_upstream(self, upstream_data_sources):
        """Add upstream data sources.

        Parameters
        ----------
        upstream_data_sources : DataSources
            Set of paths creating an upstream for recursive workflows.

        """
        request = data_sources_pb2.UpdateUpstreamRequest()
        request.upstream_data_sources.CopyFrom(upstream_data_sources._message)
        request.data_sources.CopyFrom(self._message)
        self._stub.UpdateUpstream(request)

    @property
    def result_key(self):
        """Result key used by the data sources.

        Returns
        -------
        str
           Result key.

        """
        return self._info["result_key"]

    @property
    def result_files(self):
        """List of result files contained in the data sources.

        Returns
        ----------
        list
            List of result files.
        """
        key = self.result_key
        if key == "":
            return None
        else:
            return self._info["paths"][key]

    @property
    def _info(self):
        list = self._stub.List(self._message)
        paths = {}
        for key in list.paths:
            key_paths = []
            for path in list.paths[key].paths:
                key_paths.append(path)
            paths[key] = key_paths
        out = {"result_key": list.result_key, "paths": paths}
        return out

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        str
            Description of the entity.
        """
        from ansys.dpf.core.core import _description

        return _description(self._message, self._server)

    def __del__(self):
        try:  # should silently fail
            self._stub.Delete(self._message)
        except:
            pass



"""
Server
======
Contains the directives necessary to start the DPF server.
"""
from threading import Thread
import io
import platform
import logging
import time
import os
import socket
import subprocess
import grpc
import psutil
import weakref
import atexit
import copy

from ansys import dpf
from ansys.dpf.core.misc import find_ansys, is_ubuntu
from ansys.dpf.core import errors

from ansys.dpf.core._version import __ansys_version__

MAX_PORT = 65535

LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")

# default DPF server port
DPF_DEFAULT_PORT = int(os.environ.get("DPF_PORT", 50054))
LOCALHOST = os.environ.get("DPF_IP", "127.0.0.1")


def shutdown_global_server():
    try:
        if dpf.core.SERVER != None:
            del dpf.core.SERVER
    except:
        pass


atexit.register(shutdown_global_server)


def has_local_server():
    """Check if a local DPF gRPC server has been created.

    Returns
    -------
    bool
        ``True`` when a local DPF gRPC server has been created.

    """
    return dpf.core.SERVER is not None


def _global_server():
    """Retrieve the global server if it exists.

    If the global server has not been specified, check if the user
    has specified the "DPF_START_SERVER" environment variable.  If
    ``True``, start the server locally.  If ``False``, connect to the
    existing server.
    """
    if dpf.core.SERVER is None:
        if os.environ.get("DPF_START_SERVER", "").lower() == "false":
            ip = os.environ.get("DPF_IP", LOCALHOST)
            port = int(os.environ.get("DPF_PORT", DPF_DEFAULT_PORT))
            connect_to_server(ip, port)
        else:
            start_local_server()

    return dpf.core.SERVER


def port_in_use(port, host=LOCALHOST):
    """Check if a port is in use at the given host.

    The port must actually "bind" the address. Just checking to see if a
    socket can be created is insufficient because it's possible to run into
    permission errors like: ``An attempt was made to access a socket in a way
    forbidden by its access permissions.``

    Returns
    -------
    bool
        ``True`` when the port is in use, ``False`` when free.

    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.bind((host, port))
            return False
        except:
            return True


def check_valid_ip(ip):
    """Check if a valid IP address is entered.

    This method raises an error when an invalid IP address is entered.
    """
    try:
        socket.inet_aton(ip)
    except OSError:
        raise ValueError(f'Invalid IP address "{ip}"')


def shutdown_all_session_servers():
    """Shut down all active servers created by this module."""
    from ansys.dpf.core import _server_instances

    copy_instances = copy.deepcopy(_server_instances)
    for instance in copy_instances:
        try:
            instance().shutdown()
        except Exception as e:
            print(e.args)
            pass


def start_local_server(
    ip=LOCALHOST,
    port=DPF_DEFAULT_PORT,
    ansys_path=None,
    as_global=True,
    load_operators=True,
):
    """Start a new local DPF server at a given port and IP address.

    This method requires Windows and ANSYS 2021 R1 or later. If ``as_global=True``, which is
    the default) the server is stored globally, replacing the one stored previously.
    Otherwise, a user must keep a handle on their server.

    Parameters
    ----------
    ip : str, optional
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    ansys_path : str, optional
        Root path for the Ansys installation directory. For example, ``"/ansys_inc/v212/"``.
        The default is the latest Ansys installation.
    as_global : bool, optional
        Global variable that stores the IP address and port for the DPF
        module. All DPF objects created in this Python session will
        use this IP and port. The default is ``True``.
    load_operators : bool, optional
        Whether to automatically load the math operators. The default is ``True``.

    Returns
    -------
    server : server.DpfServer
    """
    if ansys_path is None:
        ansys_path = os.environ.get("AWP_ROOT" + __ansys_version__, find_ansys())
    if ansys_path is None:
        raise ValueError(
            "Unable to automatically locate the Ansys path.  "
            "Manually enter one when starting the server or set it "
            'as the environment variable "ANSYS_PATH"'
        )

    # verify path exists
    if not os.path.isdir(ansys_path):
        raise NotADirectoryError(f'Invalid Ansys path "{ansys_path}"')

    # parse the version to an int and check for supported
    try:
        ver = int(ansys_path[-3:])
        if ver < 211:
            raise errors.InvalidANSYSVersionError(f"Ansys v{ver} does not support DPF")
        if ver == 211 and is_ubuntu():
            raise OSError("DPF on v211 does not support Ubuntu")
    except ValueError:
        pass

    # avoid using any ports in use from existing servers
    used_ports = []
    if dpf.core._server_instances:
        for srv in dpf.core._server_instances:
            if srv():
                used_ports.append(srv().port)

    while port in used_ports:
        port += 1

    # verify port is free
    while port_in_use(port):
        port += 1

    server = None
    n_attempts = 10
    for _ in range(n_attempts):
        try:
            server = DpfServer(
                ansys_path, ip, port, as_global=as_global, load_operators=load_operators
            )
            break
        except errors.InvalidPortError:  # allow socket in use errors
            port += 1

    if server is None:
        raise OSError(
            f"Unable to launch the server after {n_attempts} attempts.  "
            "Check the following path:\n{ansys_path}\n\n"
            "or attempt to use a different port"
        )

    dpf.core._server_instances.append(weakref.ref(server))
    return server


def connect_to_server(ip=LOCALHOST, port=DPF_DEFAULT_PORT, as_global=True, timeout=5):
    """Connect to an existing DPF server.

    This method sets the global default channel that is then used for the
    duration of the DPF sesssion.

    Parameters
    ----------
    ip : str
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    as_global : bool, optional
        Global variable that stores the IP address and port for the DPF
        module. All DPF objects created in this Python session will
        use this IP and port. The default is ``True``.
    timeout : float, optional
        Maximum number of seconds for the initalization attempt.
        The default is ``10``. Once the specified number of seconds
        passes, the connection fails.

    Examples
    --------

    >>> from ansys.dpf import core as dpf

    Create a server.

    >>> #server = dpf.start_local_server(ip = '127.0.0.1')
    >>> #port = server.port

    Connect to a remote server at a non-default port.

    >>> #specified_server = dpf.connect_to_server('127.0.0.1', port, as_global=False)

    Connect to the localhost at the default port.

    >>> #unspecified_server = dpf.connect_to_server(as_global=False)

    """
    server = DpfServer(ip=ip, port=port, as_global=as_global, launch_server=False)
    dpf.core._server_instances.append(weakref.ref(server))
    return server


class DpfServer:
    """Provides an instance of the DPF server.

    Parameters
    -----------
    server_bin : str
        Path for the DPF executable.
    ip : str
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    timeout : float, optional
        Maximum number of seconds for the initalization attempt.
        The default is ``10``. Once the specified number of seconds
        passes, the connection fails.
    as_global : bool, optional
        Global variable that stores the IP address and port for the DPF
        module. All DPF objects created in this Python session will
        use this IP and port. The default is ``True``.
    load_operators : bool, optional
        Whether to automatically load the math operators. The default
        is ``True``.
    launch_server : bool, optional
        Whether to launch the server on Windows.
    """

    def __init__(
        self,
        ansys_path="",
        ip=LOCALHOST,
        port=DPF_DEFAULT_PORT,
        timeout=10,
        as_global=True,
        load_operators=True,
        launch_server=True,
    ):
        """Start the DPF server."""
        # check valid ip and port
        check_valid_ip(ip)
        if not isinstance(port, int):
            raise ValueError("Port must be an integer")

        if os.name == "posix" and "ubuntu" in platform.platform().lower():
            raise OSError("DPF does not support Ubuntu")
        elif launch_server:
            launch_dpf(ansys_path, ip, port)

        self.channel = grpc.insecure_channel("%s:%d" % (ip, port))

        if launch_server is False:
            state = grpc.channel_ready_future(self.channel)
            # verify connection has matured
            tstart = time.time()
            while ((time.time() - tstart) < timeout) and not state._matured:
                time.sleep(0.01)

            if not state._matured:
                raise TimeoutError(
                    f"Failed to connect to {ip}:{port} in {timeout} seconds"
                )

            LOG.debug("Established connection to DPF gRPC")

        # assign to global channel when requested
        if as_global:
            dpf.core.SERVER = self

        # TODO: add to PIDs ...

        # store port and ip for later reference
        self.live = True
        self.ansys_path = ansys_path
        self._input_ip = ip
        self._input_port = port
        self._own_process = launch_server

    @property
    def _base_service(self):
        if not hasattr(self, "__base_service"):
            from ansys.dpf.core.core import BaseService

            self.__base_service = BaseService(self, timeout=1)
        return self.__base_service

    @property
    def info(self):
        """Server information.

        Returns
        -------
        info : dictionary
            Dictionary with server information, including ``"server_ip"``,
            ``"server_port"``, ``"server_process_id"``, and
            ``"server_version"`` keys.
        """
        return self._base_service.server_info

    @property
    def ip(self):
        """IP address of the server.

        Returns
        -------
        ip : str
        """
        try:
            return self._base_service.server_info["server_ip"]
        except:
            return ""

    @property
    def port(self):
        """Port of the server.

        Returns
        -------
        port : int
        """
        try:
            return self._base_service.server_info["server_port"]
        except:
            return 0

    @property
    def version(self):
        """Version of the server.

        Returns
        -------
        version : str
        """
        return self._base_service.server_info["server_version"]

    def __str__(self):
        return f"DPF Server: {self.info}"

    def shutdown(self):
        if self._own_process and self.live and self._base_service:
            self._base_service._prepare_shutdown()
            p = psutil.Process(self._base_service.server_info["server_process_id"])
            p.kill()
            time.sleep(0.1)
            self.live = False
            try:
                if id(dpf.core.SERVER) == id(self):
                    dpf.core.SERVER = None
            except:
                pass

            try:
                for i, server in enumerate(dpf.core._server_instances):
                    if server() == self:
                        dpf.core._server_instances.remove(server)
            except:
                pass

    def __eq__(self, other_server):
        """Return true, if the ip and the port are equals"""
        if isinstance(other_server, DpfServer):
            return self.ip == other_server.ip and self.port == other_server.port
        return False

    def __ne__(self, other_server):
        """Return true, if the ip or the port are different"""
        return not self.__eq__(other_server)

    def __del__(self):
        try:
            self.shutdown()
        except:
            pass

    def check_version(self, required_version, msg=None):
        """Check if the server version matches with a required version.

        Parameters
        ----------
        required_version : str
            Required version to compare with the server version.
        msg : str, optional
            Message for the raised exception if version requirements do not match.

        Raises
        ------
        dpf_errors : errors
            errors.DpfVersionNotSupported is raised if failure.

        Returns
        -------
        bool
            ``True`` if the server version meets the requirement.
        """
        from ansys.dpf.core.check_version import server_meet_version_and_raise

        return server_meet_version_and_raise(required_version, self, msg)


def launch_dpf(ansys_path, ip=LOCALHOST, port=DPF_DEFAULT_PORT, timeout=10):
    """Launch Ansys DPF.

    Parameters
    ----------
    ansys_path : str, optional
        Root path for the Ansys installation directory. For example, ``"/ansys_inc/v212/"``.
        The default is the latest Ansys installation.
    ip : str, optional
        IP address of the remote or local instance to connect to. The
        default is ``"LOCALHOST"``.
    port : int
        Port to connect to the remote instance on. The default is
        ``"DPF_DEFAULT_PORT"``, which is 50054.
    timeout : float, optional
        Maximum number of seconds for the initalization attempt.
        The default is ``10``. Once the specified number of seconds
        passes, the connection fails.

    Returns
    -------
    process : subprocess.Popen
        DPF Process.
    """
    if os.name == "nt":
        run_cmd = f"Ans.Dpf.Grpc.bat --address {ip} --port {port}"
        path_in_install = "aisol/bin/winx64"
    else:
        run_cmd = ["./Ans.Dpf.Grpc.sh", f"--address {ip}", f"--port {port}"]
        path_in_install = "aisol/bin/linx64"

    # verify ansys path is valid
    if os.path.isdir(f"{ansys_path}/{path_in_install}"):
        dpf_run_dir = f"{ansys_path}/{path_in_install}"
    else:
        dpf_run_dir = f"{ansys_path}"
    if not os.path.isdir(dpf_run_dir):
        raise NotADirectoryError(
            f'Invalid ansys path at "{ansys_path}".  '
            "Unable to locate the directory containing DPF at "
            f'"{dpf_run_dir}"'
        )

    old_dir = os.getcwd()
    os.chdir(dpf_run_dir)
    process = subprocess.Popen(run_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    os.chdir(old_dir)

    # check to see if the service started
    lines = []

    def read_stdout():
        for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
            LOG.debug(line)
            lines.append(line)

    errors = []

    def read_stderr():
        for line in io.TextIOWrapper(process.stderr, encoding="utf-8"):
            LOG.error(line)
            errors.append(line)

    # must be in the background since the process reader is blocking
    Thread(target=read_stdout, daemon=True).start()
    Thread(target=read_stderr, daemon=True).start()

    t_timeout = time.time() + timeout
    started = False
    while not started:
        started = any("server started" in line for line in lines)

        if time.time() > t_timeout:
            raise TimeoutError(f"Server did not start in {timeout} seconds")

    # verify there were no errors
    time.sleep(1)
    if errors:
        try:
            process.kill()
        except PermissionError:
            pass
        errstr = "\n".join(errors)
        if "Only one usage of each socket address" in errstr:
            raise errors.InvalidPortError(f"Port {port} in use")
        raise RuntimeError(errstr)



"""
transpose
=========
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "scoping" category
"""

class transpose(Operator):
    """Transposes the input scoping or scopings container (Elemental --> Nodal, or Nodal ---> Elemental), based on the input mesh region.

      available inputs:
        - mesh_scoping (Scoping, ScopingsContainer)
        - meshed_region (MeshedRegion, MeshesContainer)
        - inclusive (int) (optional)

      available outputs:
        - mesh_scoping (Scoping ,ScopingsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.scoping.transpose()

      >>> # Make input connections
      >>> my_mesh_scoping = dpf.Scoping()
      >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
      >>> my_meshed_region = dpf.MeshedRegion()
      >>> op.inputs.meshed_region.connect(my_meshed_region)
      >>> my_inclusive = int()
      >>> op.inputs.inclusive.connect(my_inclusive)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.scoping.transpose(mesh_scoping=my_mesh_scoping,meshed_region=my_meshed_region,inclusive=my_inclusive)

      >>> # Get output data
      >>> result_mesh_scoping = op.outputs.mesh_scoping()"""
    def __init__(self, mesh_scoping=None, meshed_region=None, inclusive=None, config=None, server=None):
        super().__init__(name="transpose_scoping", config = config, server = server)
        self._inputs = InputsTranspose(self)
        self._outputs = OutputsTranspose(self)
        if mesh_scoping !=None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if meshed_region !=None:
            self.inputs.meshed_region.connect(meshed_region)
        if inclusive !=None:
            self.inputs.inclusive.connect(inclusive)

    @staticmethod
    def _spec():
        spec = Specification(description="""Transposes the input scoping or scopings container (Elemental --> Nodal, or Nodal ---> Elemental), based on the input mesh region.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "mesh_scoping", type_names=["scoping","scopings_container"], optional=False, document="""Scoping or scopings container (the input type is the output type)"""), 
                                 1 : PinSpecification(name = "meshed_region", type_names=["meshed_region","meshes_container"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "inclusive", type_names=["int32"], optional=True, document="""if inclusive == 1 then all the elements adjacent to the nodes ids in input are added, if inclusive == 0, only the elements which have all their nodes in the scoping are included""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "mesh_scoping", type_names=["scoping","scopings_container"], optional=False, document="""Scoping or scopings container (the input type is the output type)""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "transpose_scoping")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsTranspose 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsTranspose 
        """
        return super().outputs


#internal name: transpose_scoping
#scripting name: transpose
class InputsTranspose(_Inputs):
    """Intermediate class used to connect user inputs to transpose operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.scoping.transpose()
      >>> my_mesh_scoping = dpf.Scoping()
      >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
      >>> my_meshed_region = dpf.MeshedRegion()
      >>> op.inputs.meshed_region.connect(my_meshed_region)
      >>> my_inclusive = int()
      >>> op.inputs.inclusive.connect(my_inclusive)
    """
    def __init__(self, op: Operator):
        super().__init__(transpose._spec().inputs, op)
        self._mesh_scoping = Input(transpose._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._mesh_scoping)
        self._meshed_region = Input(transpose._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._meshed_region)
        self._inclusive = Input(transpose._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._inclusive)

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator

        - pindoc: Scoping or scopings container (the input type is the output type)

        Parameters
        ----------
        my_mesh_scoping : Scoping, ScopingsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> #or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)

        """
        return self._mesh_scoping

    @property
    def meshed_region(self):
        """Allows to connect meshed_region input to the operator

        Parameters
        ----------
        my_meshed_region : MeshedRegion, MeshesContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.meshed_region.connect(my_meshed_region)
        >>> #or
        >>> op.inputs.meshed_region(my_meshed_region)

        """
        return self._meshed_region

    @property
    def inclusive(self):
        """Allows to connect inclusive input to the operator

        - pindoc: if inclusive == 1 then all the elements adjacent to the nodes ids in input are added, if inclusive == 0, only the elements which have all their nodes in the scoping are included

        Parameters
        ----------
        my_inclusive : int, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.scoping.transpose()
        >>> op.inputs.inclusive.connect(my_inclusive)
        >>> #or
        >>> op.inputs.inclusive(my_inclusive)

        """
        return self._inclusive

class OutputsTranspose(_Outputs):
    """Intermediate class used to get outputs from transpose operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.scoping.transpose()
      >>> # Connect inputs : op.inputs. ...
      >>> result_mesh_scoping = op.outputs.mesh_scoping()
    """
    def __init__(self, op: Operator):
        super().__init__(transpose._spec().outputs, op)
        self.mesh_scoping_as_scoping = Output( _modify_output_spec_with_one_type(transpose._spec().output_pin(0), "scoping"), 0, op) 
        self._outputs.append(self.mesh_scoping_as_scoping)
        self.mesh_scoping_as_scopings_container = Output( _modify_output_spec_with_one_type(transpose._spec().output_pin(0), "scopings_container"), 0, op) 
        self._outputs.append(self.mesh_scoping_as_scopings_container)




"""
minus
=====
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "math" category
"""

class minus(Operator):
    """Computes the difference of two fields. If one field's scoping has 'overall' location, then these field's values are applied on the entire other field.

      available inputs:
        - fieldA (Field, FieldsContainer)
        - fieldB (Field, FieldsContainer)

      available outputs:
        - field (Field)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.math.minus()

      >>> # Make input connections
      >>> my_fieldA = dpf.Field()
      >>> op.inputs.fieldA.connect(my_fieldA)
      >>> my_fieldB = dpf.Field()
      >>> op.inputs.fieldB.connect(my_fieldB)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.math.minus(fieldA=my_fieldA,fieldB=my_fieldB)

      >>> # Get output data
      >>> result_field = op.outputs.field()"""
    def __init__(self, fieldA=None, fieldB=None, config=None, server=None):
        super().__init__(name="minus", config = config, server = server)
        self._inputs = InputsMinus(self)
        self._outputs = OutputsMinus(self)
        if fieldA !=None:
            self.inputs.fieldA.connect(fieldA)
        if fieldB !=None:
            self.inputs.fieldB.connect(fieldB)

    @staticmethod
    def _spec():
        spec = Specification(description="""Computes the difference of two fields. If one field's scoping has 'overall' location, then these field's values are applied on the entire other field.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fieldA", type_names=["field","fields_container"], optional=False, document="""field or fields container with only one field is expected"""), 
                                 1 : PinSpecification(name = "fieldB", type_names=["field","fields_container"], optional=False, document="""field or fields container with only one field is expected""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "minus")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMinus 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsMinus 
        """
        return super().outputs


#internal name: minus
#scripting name: minus
class InputsMinus(_Inputs):
    """Intermediate class used to connect user inputs to minus operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.minus()
      >>> my_fieldA = dpf.Field()
      >>> op.inputs.fieldA.connect(my_fieldA)
      >>> my_fieldB = dpf.Field()
      >>> op.inputs.fieldB.connect(my_fieldB)
    """
    def __init__(self, op: Operator):
        super().__init__(minus._spec().inputs, op)
        self._fieldA = Input(minus._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fieldA)
        self._fieldB = Input(minus._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._fieldB)

    @property
    def fieldA(self):
        """Allows to connect fieldA input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_fieldA : Field, FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.minus()
        >>> op.inputs.fieldA.connect(my_fieldA)
        >>> #or
        >>> op.inputs.fieldA(my_fieldA)

        """
        return self._fieldA

    @property
    def fieldB(self):
        """Allows to connect fieldB input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_fieldB : Field, FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.minus()
        >>> op.inputs.fieldB.connect(my_fieldB)
        >>> #or
        >>> op.inputs.fieldB(my_fieldB)

        """
        return self._fieldB

class OutputsMinus(_Outputs):
    """Intermediate class used to get outputs from minus operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.minus()
      >>> # Connect inputs : op.inputs. ...
      >>> result_field = op.outputs.field()
    """
    def __init__(self, op: Operator):
        super().__init__(minus._spec().outputs, op)
        self._field = Output(minus._spec().output_pin(0), 0, op) 
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator


        Returns
        ----------
        my_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.minus()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field() 
        """
        return self._field




"""
Element Descriptor
==================
"""


class ElementDescriptor:
    """Describes an element.

    Parameters
    ----------
    element_id: int

    description: str
        Specifies the element geometry and integration order.

    name: str

    shape: str, optional
        Can be ``"solid"``, ``"shell"``, or ``"beam"``. The default is ``None``.

    n_corner_nodes: int, optional
        The default is ``None``.
    n_mid_nodes: int, optional
        The default is ``None``.
    n_nodes: int, optional
        The default is ``None``.
    is_solid: bool, optional
        Whether the element is a solid. The default is ``None``.
    is_shell: bool, optional
        Whether the element is a shell. The default is ``None``.
    is_beam: bool, optional
        Whether the element is a beam. The default is ``None``.
    is_quadratic: bool, optional
        Whether the element is a quadratic. The default is ``None``.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> descriptor = dpf.ElementDescriptor(10, "Linear 4-nodes Tetrahedron", "tet4", "solid", 4, 0, 4, True, False, False, False)

    """  # noqa: E501

    def __init__(
        self,
        enum_id,
        description,
        name,
        shape=None,
        n_corner_nodes=None,
        n_mid_nodes=None,
        n_nodes=None,
        is_solid=None,
        is_shell=None,
        is_beam=None,
        is_quadratic=None,
    ):
        """Constructor of ElementDescriptor."""
        self.enum_id = enum_id
        self.description = description
        self.name = name
        self.n_corner_nodes = n_corner_nodes
        self.n_mid_nodes = n_mid_nodes
        self.n_nodes = n_nodes
        self.shape = shape
        if self.shape is None:
            self.shape = "unknown_shape"
        self.is_solid = is_solid
        self.is_shell = is_shell
        self.is_beam = is_beam
        self.is_quadratic = is_quadratic

    def __str__(self):
        lines = []
        lines.append("Element descriptor")
        lines.append("-" * 18)
        lines.append(f"Enum id (dpf.element_types): {self.enum_id}")
        lines.append(f"Element description: {self.description}")
        lines.append(f"Element name (short): {self.name}")
        lines.append(f"Element shape: {self.shape}")
        lines.append(f"Number of corner nodes: {self.n_corner_nodes}")
        lines.append(f"Number of mid-side nodes: {self.n_mid_nodes}")
        lines.append(f"Total number of nodes: {self.n_nodes}")
        lines.append(f"Quadratic element: {self.is_quadratic}")
        return "\n".join(lines)



"""
fields_container_factory
========================

Contains functions to simplify creating a fields container.
"""

from ansys.dpf.core import FieldsContainer, TimeFreqSupport
from ansys.dpf.core import errors as dpf_errors
from ansys.dpf.core import fields_factory
from ansys.dpf.core.common import locations


def over_time_freq_fields_container(fields, time_freq_unit=None, server=None):
    """Create a fields container with one field by time set.

    This method can also set the time frequency support with the correct unit
    if needed.

    Parameters
    ----------
    fields : Dictionary(time_int_key : Field) or list of Field
        Dictionary of field entities to add to the fields container.
    time_freq_unit : str, optional
        Unit of the time frequency support, which is taken into
        account if the fields attribute has a dictionary type. The
        default is ``None``.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    fields_container : FieldsContainer
        Fields container containing one field by time step.

    Examples
    --------
    Create a fields container from scratch based on time labels.

    >>> from ansys.dpf import core as dpf
    >>> field1 = dpf.Field()
    >>> field2 = dpf.Field()
    >>> from ansys.dpf.core import fields_container_factory
    >>> my_fc = fields_container_factory.over_time_freq_fields_container([ field1, field2 ])

    """
    if not isinstance(fields, dict) and not isinstance(fields, list):
        raise dpf_errors.InvalidTypeError("dictionary/list", "fields")
    fc = FieldsContainer(server=server)
    fc.labels = {"time"}
    i = 0
    # dict case
    if isinstance(fields, dict):
        time_freq = []
        for field_key in fields:
            fc.add_field({"time": i + 1}, fields[field_key])
            time_freq.append(field_key)
            i += 1
        time_freq_field = fields_factory.create_scalar_field(
            len(fields), location=locations.time_freq, server=server
        )
        time_freq_field.append(time_freq, 1)
        time_freq_field.unit = time_freq_unit
        time_freq_support = TimeFreqSupport(server=server)
        time_freq_support.time_frequencies = time_freq_field
        fc.time_freq_support = time_freq_support
    # list case
    elif isinstance(fields, list):
        for field in fields:
            fc.add_field({"time": i + 1}, field)
            i += 1
    return fc


def over_time_freq_complex_fields_container(
    real_fields, imaginary_fields, time_freq_unit=None, server=None
):
    """Create a fields container with two fields (real and imaginary) by time set.

    If the inputs for the fields are dictionaries, this method sets the time frequency
    support with the correct unit if needed.

    Parameters
    ----------
    real_fields : Dictionary(time_int_key : Field) or list of Field
        Dictionary or list of field entities to add to the fields container.
    imaginary_fields : Dictionary(time_int_key : Field) or list of Field
        Dictionary or list of field entities to add to the fields container.
    time_freq_unit : str , optional
        Unit of the time frequency support, which is taken into account if
        the field's attribute has a dictionary type.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    fields_container : FieldsContainer
        Fields container containing two fields (real and imaginary) by time step.
    """
    if not isinstance(real_fields, dict) and not isinstance(real_fields, list):
        raise dpf_errors.InvalidTypeError("dictionary/list", "real_fields")
    if not isinstance(imaginary_fields, dict) and not isinstance(
        imaginary_fields, list
    ):
        raise dpf_errors.InvalidTypeError("dictionary/list", "imaginary_fields")

    errorString = (
        "Both real_fields and imaginary_fields must have the same type (list or dict)"
    )
    if isinstance(real_fields, dict):
        if not isinstance(imaginary_fields, dict):
            raise dpf_errors.DpfValueError(errorString)
    elif isinstance(real_fields, list):
        if not isinstance(imaginary_fields, list):
            raise dpf_errors.DpfValueError(errorString)

    fc = FieldsContainer(server=server)
    fc.labels = ["time", "complex"]
    i = 0
    # dict case
    if isinstance(real_fields, dict):
        time_freq = []
        for field_key in real_fields:
            fc.add_field({"time": i + 1, "complex": 0}, real_fields[field_key])
            time_freq.append(field_key)
            i += 1
        i = 0
        im_time_freq = []
        for field_key in imaginary_fields:
            fc.add_field({"time": i + 1, "complex": 1}, imaginary_fields[field_key])
            im_time_freq.append(field_key)
            i += 1
        time_freq_field = fields_factory.create_scalar_field(
            len(real_fields), locations.time_freq, server=server
        )
        time_freq_field.append(time_freq, 1)
        time_freq_field.unit = time_freq_unit
        im_time_freq_field = fields_factory.create_scalar_field(
            len(imaginary_fields), locations.time_freq, server=server
        )
        im_time_freq_field.append(im_time_freq, 1)
        im_time_freq_field.unit = time_freq_unit
        time_freq_support = TimeFreqSupport(server=server)
        time_freq_support.time_frequencies = time_freq_field
        time_freq_support.complex_frequencies = im_time_freq_field
        fc.time_freq_support = time_freq_support
    # list case
    if isinstance(real_fields, list):
        for field in real_fields:
            fc.add_field({"time": i + 1, "complex": 0}, field)
            i += 1
        i = 0
        for field in imaginary_fields:
            fc.add_field({"time": i + 1, "complex": 1}, field)
            i += 1
    return fc


def complex_fields_container(real_field, imaginary_field, server=None):
    """Create a fields container with two fields (real and imaginary) and only one time set.

    Parameters
    ----------
    real_fields : Field
        Real :class:`ansys.dpf.core.Field` entity to add to the fields container.
    imaginary_fields : Field
        Imaginary :class:`ansys.dpf.core.Field` entity to add to the fields container.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    fields_container : FieldsContainer
        Fields container with two fields (real and imaginary).
    """
    fc = FieldsContainer(server=server)
    fc.labels = ["complex"]
    fc.add_field({"complex": 0}, real_field)
    fc.add_field({"complex": 1}, imaginary_field)
    return fc



import numpy as np
from vtk import (
    VTK_VERTEX,
    VTK_LINE,
    VTK_TRIANGLE,
    VTK_QUAD,
    VTK_TETRA,
    VTK_HEXAHEDRON,
    VTK_WEDGE,
    VTK_PYRAMID,
    VTK_QUADRATIC_EDGE,
    VTK_QUADRATIC_TRIANGLE,
    VTK_QUADRATIC_QUAD,
    VTK_QUADRATIC_TETRA,
    VTK_QUADRATIC_HEXAHEDRON,
    VTK_QUADRATIC_PYRAMID,
    vtkVersion,
)
import pyvista as pv

VTK9 = vtkVersion().GetVTKMajorVersion() >= 9

# Maps dpf cell sizes (based on array order) to the number of nodes per cell
SIZE_MAPPING = np.array(
    [
        10,  # kAnsTet10
        20,  # kAnsHex20
        15,  # kAnsWedge15
        13,  # kAnsPyramid13
        6,  # kAnsTri6
        6,  # kAnsTriShell6
        8,  # kAnsQuad8
        8,  # kAnsQuadShell8
        3,  # kAnsLine3
        1,  # kAnsPoint1
        4,  # kAnsTet4
        8,  # kAnsHex8
        6,  # kAnsWedge6
        5,  # kAnsPyramid5
        3,  # kAnsTri3
        3,  # kAnsTriShell3
        4,  # kAnsQuad4
        4,  # kAnsQuadShell4
        2,  # kAnsLine2
        0,  # kAnsNumElementTypes
        0,  # kAnsUnknown
        0,  # kAnsEMagLine
        0,  # kAnsEMagArc
        0,  # kAnsEMagCircle
        3,  # kAnsSurface3
        4,  # kAnsSurface4
        6,  # kAnsSurface6
        8,  # kAnsSurface8
        2,  # kAnsEdge2
        3,  # kAnsEdge3
        3,  # kAnsBeam3
        4,
    ]
)  # kAnsBeam4


# DPF --> VTK mapping
# any cells not mapped will be empty cells
VTK_MAPPING = np.array(
    [
        VTK_QUADRATIC_TETRA,  # kAnsTet10 = 0,
        VTK_QUADRATIC_HEXAHEDRON,  # kAnsHex20 = 1,
        # VTK_QUADRATIC_WEDGE,  # kAnsWedge15 = 2,
        VTK_WEDGE,  # kAnsWedge15 = 2,
        VTK_QUADRATIC_PYRAMID,  # kAnsPyramid13 = 3,
        VTK_QUADRATIC_TRIANGLE,  # kAnsTri6 = 4,
        VTK_QUADRATIC_TRIANGLE,  # kAnsTriShell6 = 5,
        VTK_QUADRATIC_QUAD,  # kAnsQuad8 = 6,
        VTK_QUADRATIC_QUAD,  # kAnsQuadShell8 = 7,
        VTK_QUADRATIC_EDGE,  # kAnsLine3 = 8,
        VTK_VERTEX,  # kAnsPoint1 = 9,
        VTK_TETRA,  # kAnsTet4 = 10,
        VTK_HEXAHEDRON,  # kAnsHex8 = 11,
        VTK_WEDGE,  # kAnsWedge6 = 12,
        VTK_PYRAMID,  # kAnsPyramid5 = 13,
        VTK_TRIANGLE,  # kAnsTri3 = 14,
        VTK_TRIANGLE,  # kAnsTriShell3 = 15,
        VTK_QUAD,  # kAnsQuad4 = 16,
        VTK_QUAD,  # kAnsQuadShell4 = 17,
        VTK_LINE,  # kAnsLine2 = 18,
        0,  # kAnsNumElementTypes = 19,
        0,  # kAnsUnknown = 20,
        0,  # kAnsEMagLine = 21,
        0,  # kAnsEMagArc = 22,
        0,  # kAnsEMagCircle = 23,
        VTK_TRIANGLE,  # kAnsSurface3 = 24,
        VTK_QUAD,  # kAnsSurface4 = 25,
        VTK_QUADRATIC_TRIANGLE,  # kAnsSurface6 = 26,
        VTK_QUADRATIC_QUAD,  # kAnsSurface8 = 27,
        0,  # kAnsEdge2 = 28,
        0,  # kAnsEdge3 = 29,
        0,  # kAnsBeam3 = 30,
        0,
    ]
)  # kAnsBeam4 = 31,


# map all cells to linear
VTK_LINEAR_MAPPING = np.array(
    [
        VTK_TETRA,  # kAnsTet10 = 0,
        VTK_HEXAHEDRON,  # kAnsHex20 = 1,
        VTK_WEDGE,  # kAnsWedge15 = 2,
        VTK_PYRAMID,  # kAnsPyramid13 = 3,
        VTK_TRIANGLE,  # kAnsTri6 = 4,
        VTK_TRIANGLE,  # kAnsTriShell6 = 5,
        VTK_QUAD,  # kAnsQuad8 = 6,
        VTK_QUAD,  # kAnsQuadShell8 = 7,
        VTK_LINE,  # kAnsLine3 = 8,
        VTK_VERTEX,  # kAnsPoint1 = 9,
        VTK_TETRA,  # kAnsTet4 = 10,
        VTK_HEXAHEDRON,  # kAnsHex8 = 11,
        VTK_WEDGE,  # kAnsWedge6 = 12,
        VTK_PYRAMID,  # kAnsPyramid5 = 13,
        VTK_TRIANGLE,  # kAnsTri3 = 14,
        VTK_TRIANGLE,  # kAnsTriShell3 = 15,
        VTK_QUAD,  # kAnsQuad4 = 16,
        VTK_QUAD,  # kAnsQuadShell4 = 17,
        VTK_LINE,  # kAnsLine2 = 18,
        0,  # kAnsNumElementTypes = 19,
        0,  # kAnsUnknown = 20,
        0,  # kAnsEMagLine = 21,
        0,  # kAnsEMagArc = 22,
        0,  # kAnsEMagCircle = 23,
        VTK_TRIANGLE,  # kAnsSurface3 = 24,
        VTK_QUAD,  # kAnsSurface4 = 25,
        VTK_QUADRATIC_TRIANGLE,  # kAnsSurface6 = 26,
        VTK_QUADRATIC_QUAD,  # kAnsSurface8 = 27,
        0,  # kAnsEdge2 = 28,
        0,  # kAnsEdge3 = 29,
        0,  # kAnsBeam3 = 30,
        0,
    ]
)  # kAnsBeam4 = 31,


def dpf_mesh_to_vtk(nodes, etypes, connectivity, as_linear=True):
    """Return a pyvista unstructured grid given DPF node and element
    definitions.

    Parameters
    ----------
    nodes : np.ndarray
        Numpy array containing the nodes of the mesh.

    etypes : np.ndarray
        ANSYS DPF element types.

    connectivity : np.ndarray
        Array containing the nodes used by each element.

    Returns
    -------
    grid : pyvista.UnstructuredGrid
        Unstructred grid of the DPF mesh.
    """
    # could make this more efficient in C...
    elem_size = SIZE_MAPPING[etypes]
    insert_ind = np.cumsum(elem_size)
    insert_ind = np.hstack(([0], insert_ind))[:-1]

    # TODO: Investigate why connectivity can be -1
    nullmask = connectivity == -1
    connectivity[nullmask] = 0
    if nullmask.any():
        nodes[0] = np.nan

    # partition cells in vtk format
    cells = np.insert(connectivity, insert_ind, elem_size)

    def compute_offset():
        """Return the starting point of a cell in the cells array"""
        return insert_ind + np.arange(insert_ind.size)

    # convert kAns to VTK cell type
    offset = None
    if as_linear:
        vtk_cell_type = VTK_LINEAR_MAPPING[etypes]

        # visualization bug within VTK with quadratic surf cells
        ansquad8_mask = etypes == 6
        if np.any(ansquad8_mask):  # kAnsQuad8

            # simply copy the edge node indices to the midside points
            offset = compute_offset()
            cell_pos = offset[ansquad8_mask]
            cells[cell_pos + 5] = cells[cell_pos + 1]
            cells[cell_pos + 6] = cells[cell_pos + 2]
            cells[cell_pos + 7] = cells[cell_pos + 3]
            cells[cell_pos + 8] = cells[cell_pos + 4]

        anstri6_mask = etypes == 4  # kAnsTri6 = 4
        if np.any(anstri6_mask):
            if offset is None:
                offset = compute_offset()
            cell_pos = offset[anstri6_mask]
            cells[cell_pos + 4] = cells[cell_pos + 1]
            cells[cell_pos + 5] = cells[cell_pos + 2]
            cells[cell_pos + 6] = cells[cell_pos + 3]

    else:
        vtk_cell_type = VTK_MAPPING[etypes]

    # different treatment depending on the version of vtk
    if VTK9:
        # compute offset array when < VTK v9
        return pv.UnstructuredGrid(cells, vtk_cell_type, nodes)

    # might be computed when checking for VTK quadratic bug
    if offset is None:
        offset = compute_offset()

    return pv.UnstructuredGrid(offset, cells, vtk_cell_type, nodes)



"""
.. _ref_results:

Results
========
This module contains the Results and Result classes that are created by the model
to easily access results in result files."""
import functools


from ansys.dpf.core import Operator
from ansys.dpf.core import errors
from ansys.dpf.core.scoping import Scoping
from ansys.dpf.core.custom_fields_container import (
    ElShapeFieldsContainer,
    BodyFieldsContainer,
)


class Results:
    """Organizes the results from DPF into accessible methods.

    All the available results are dynamically created base on the model's class:'ResultInfo' class.

    Attributes
    ----------
    displacement : Result
        Result provider helper wrapping the regular displacement operator.
        With this wrapper, time and mesh scopings can easily
        be customized.

        Examples
        --------
        Create a displacement result from the model and choose its time and mesh scopings.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement.on_last_time_freq.on_named_selection("_CONSTRAINEDNODES")
        >>> last_time_disp = disp.eval()

    stress : Result
        Result provider helper wrapping the regular stress operator.
        With this wrapper, time and mesh scopings, location, and more
        can easily be customized.

        Examples
        --------
        Create a stress result from the model and choose its time and mesh scopings.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> stress = model.results.stress.on_last_time_freq.on_named_selection('_CONSTRAINEDNODES')
        >>> last_time_stress = stress.eval()

    .... all other results : Result
        Result provider helper wrapping all types of providers available for a
        given result file.

        Examples
        --------
        Create a stress result from the model and choose its time and mesh scopings.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.electric_therm)
        >>> v = model.results.electric_potential
        >>> rf = model.results.reaction_force
        >>> dissip = model.results.thermal_dissipation_energy

    Examples
    --------
    Extract the result object from a model.

    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.simple_bar)
    >>> results = model.results # printable object

    Access the displacement at all times.

    >>> from ansys.dpf.core import Model
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = Model(transient)
    >>> displacements = model.results.displacement.on_all_time_freqs.eval()

    """  # noqa: E501

    def __init__(self, model):
        self.__class__ = type(Results.__name__ + str(id(self)), (Results,), {})

        self._result_info = model.metadata.result_info
        self._model = model
        self._connect_operators()

    def __result__(self, result_type, *args):
        return Result(self._model, result_type)

    def _connect_operators(self):
        """Dynamically add operators for results.

        The new operator's subresults are connected to the model's
        streams.

        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = Model(transient)
        >>> disp_operator = model.results.displacement()
        >>> stress_operator = model.results.stress()
        >>> disp_x = model.results.displacement().X()
        >>> disp_y = model.results.displacement().Y()
        >>> disp_z = model.results.displacement().Z()

        """
        if self._result_info is None:
            return
        # dynamically add function based on input type
        self._op_map_rev = {}
        for result_type in self._result_info:
            try:
                doc = Operator(
                    result_type.operator_name, server=self._model._server
                ).__str__()
                bound_method = self.__result__
                method2 = functools.partial(bound_method, result_type)
                setattr(self.__class__, result_type.name, property(method2, doc=doc))

                self._op_map_rev[result_type.name] = result_type.name
            except errors.DPFServerException:
                pass
            except Exception as e:
                print(result_type.name)
                raise e

    def __str__(self):
        return str(self._result_info)

    def __iter__(self):
        for key in self._op_map_rev:
            yield self.__class__.__dict__[key].fget()

    def __getitem__(self, val):
        n = 0
        for key in self._op_map_rev:
            if n == val:
                return self.__class__.__dict__[key].fget()
            n += 1

    def __len__(self):
        return len(self._op_map_rev)


class Result:
    """Helps with using DPF's result providers.

    This class helps to connect common inputs to the operator and
    recover its fields container output. 'Result' is created by the model.

    Examples
    --------
    Create a displacement result from the model and choose its time and mesh scopings

    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.msup_transient)
    >>> disp = model.results.displacement.on_last_time_freq.on_named_selection('_CONSTRAINEDNODES')
    >>> last_time_disp = disp.eval()

    Create a stress result from the model and split the result by element shapes (solid,
    shell, and beam).

    >>> model = dpf.Model(examples.download_all_kinds_of_complexity())
    >>> stress = model.results.stress
    >>> stress_split = stress.split_by_shape.eval()
    >>> solid_stress = stress_split.solid_field()

    Create a strain result from the model on all time sets and recover the
    operator to connect it to other operators.

    >>> model = dpf.Model(examples.msup_transient)
    >>> strain = model.results.elastic_strain.on_all_time_freqs()
    >>> eqv = dpf.operators.invariant.von_mises_eqv_fc(strain)
    >>> strain_eqv = eqv.outputs.fields_container()

    """

    def __init__(self, model, result_info):
        self._model = model
        self._time_scoping = None
        self._mesh_scoping = None
        self._location = None
        self._result_info = result_info
        self._specific_fc_type = None
        from ansys.dpf.core import operators

        try:
            # create the operator to read its documentation
            # if the operator doesn't exist, the method will not be added
            doc = Operator(
                self._result_info.operator_name, server=self._model._server
            ).__str__()
            self.__doc__ = doc
            if hasattr(operators, "result") and hasattr(
                operators.result, self._result_info.name
            ):
                self._operator = getattr(operators.result, self._result_info.name)(
                    server=self._model._server
                )
            else:
                self._operator = Operator(
                    self._result_info.operator_name, server=self._model._server
                )
            self._operator._add_sub_res_operators(self._result_info.sub_results)
            self._model.__connect_op__(self._operator)
        except errors.DPFServerException:
            pass
        except Exception as e:
            print(self._result_info.name)
            raise e

    def __call__(self, time_scoping=None, mesh_scoping=None):
        op = self._operator
        if time_scoping:
            op.inputs.time_scoping(time_scoping)
        elif self._time_scoping:
            op.inputs.time_scoping(self._time_scoping)

        if mesh_scoping:
            op.inputs.mesh_scoping(mesh_scoping)
        elif self._mesh_scoping:
            op.inputs.mesh_scoping(self._mesh_scoping)

        if self._location:
            op.inputs.requested_location(self._location)

        return op

    def eval(self):
        """Evaluate the result provider with the previously specified
        inputs and return the result fields container.

        Returns
        -------
        fields_container : FieldsContainer, ElShapeFieldsContainer, BodyFieldsContainer
            If ``split_by_body`` is used, a ``BodyFieldsContainer`` is returned.
            if ``split_by_shape`` is used, an ``ElShapeFieldsContainer`` is returned.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> fc = disp.on_all_time_freqs.eval()

        """
        fc = self.__call__().outputs.fields_container()
        if self._specific_fc_type == "shape":
            fc = ElShapeFieldsContainer(fields_container=fc, server=fc._server)
        elif self._specific_fc_type == "body":
            fc = BodyFieldsContainer(fields_container=fc, server=fc._server)
        return fc

    @property
    def on_all_time_freqs(self):
        """Sets the time scoping to all the time frequencies available in the time frequency support.

        Returns
        -------
        self : Result

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> disp.on_all_time_freqs.eval().get_label_scoping("time").ids
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]

        """
        self._time_scoping = list(
            range(1, len(self._model.metadata.time_freq_support.time_frequencies) + 1)
        )
        return self

    @property
    def on_first_time_freq(self):
        """Sets the time scoping to the first time frequency available in the time frequency support.

        Returns
        -------
        self : Result

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> disp.on_first_time_freq.eval().get_label_scoping("time").ids
        [1]

        """
        self._time_scoping = 1
        return self

    @property
    def on_last_time_freq(self):
        """Sets the time scoping to the last time frequency available in the time frequency support.

        Returns
        -------
        self : Result

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> disp.on_last_time_freq.eval().get_label_scoping("time").ids
        [20]

        """
        self._time_scoping = len(
            self._model.metadata.time_freq_support.time_frequencies
        )
        return self

    def on_time_scoping(self, time_scoping):
        """Sets the time scoping to a given one.

        Parameters
        ----------
        time_scoping :  float, list[float], int, list[int], Scoping
            One or more times or frequencies.

        Returns
        -------
        self : Result

        Examples
        --------
        Choose time sets.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> stress = model.results.stress
        >>> fc = stress.on_time_scoping([1,2,3,19]).eval()
        >>> len(fc)
        4

        Choose times. If the times chosen are not in the time frequency support,
        results are extrapolated.

        >>> fc = stress.on_time_scoping([0.115,0.125]).eval()
        >>> len(fc)
        2
        >>> fc.time_freq_support.time_frequencies.data
        array([0.115, 0.125])

        """
        self._time_scoping = time_scoping
        return self

    def on_named_selection(self, named_selection):
        """Set the mesh scoping to a given named selection.

        Parameters
        ----------
        named_selection : str
            Name of the named selection or component in upper case.

        Returns
        -------
        self : Result

        Examples
        --------
        Add a requested location to the average result on the nodes.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> stress = model.results.stress
        >>> fc = stress.on_first_time_freq.on_named_selection('_CONSTRAINEDNODES').eval()
        >>> len(fc[0].scoping)
        40

        """

        self._mesh_scoping = self._model.metadata.named_selection(named_selection)
        return self

    @property
    def split_by_body(self):
        """Set the mesh scoping to a scopings container where each scoping is a body.

        Returns
        -------
        self : Result

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.download_all_kinds_of_complexity())
        >>> disp = model.results.displacement
        >>> fc_disp = disp.split_by_body.eval()
        >>> len(fc_disp)
        11
        >>> fc_disp.get_mat_scoping().ids
        [1, 5, 6, 10, 2, 7, 8, 13, 4, 12, 15]
        >>> disp_mat_10 = fc_disp.get_field_by_mat_id(10)

        """
        self._specific_fc_type = "body"
        return self._add_split_on_property_type("mat")

    @property
    def split_by_shape(self):
        """Set the mesh scoping to a scopings container where each scoping is an element shape.
        The evaluated fields container will have one field on 'solid',
        one on 'shell', one on 'beam' and one on 'unknown_shape'.

        Returns
        -------
        self : Result

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.download_all_kinds_of_complexity())
        >>> disp = model.results.displacement
        >>> fc_disp = disp.split_by_shape.eval()
        >>> len(fc_disp)
        4

        >>> shell_disp = fc_disp.shell_field()
        >>> solid_disp = fc_disp.solid_field()

        """
        self._specific_fc_type = "shape"
        return self._add_split_on_property_type("elshape")

    def _add_split_on_property_type(self, prop):
        previous_mesh_scoping = self._mesh_scoping
        from ansys.dpf.core import operators

        if hasattr(operators, "scoping") and hasattr(
            operators.scoping, "split_on_property_type"
        ):
            self._mesh_scoping = operators.scoping.split_on_property_type()
        else:
            self._mesh_scoping = Operator("scoping::by_property")

        self._mesh_scoping.inputs.requested_location(
            self._result_info.native_scoping_location
        )
        self._mesh_scoping.inputs.mesh(self._model.metadata.mesh_provider)
        self._mesh_scoping.inputs.label1(prop)
        if previous_mesh_scoping:
            try:
                self._mesh_scoping.inputs.mesh_scoping(previous_mesh_scoping)
            except:
                pass
        return self

    def on_mesh_scoping(self, mesh_scoping):
        """Set the mesh scoping to a given mesh scoping.

        Parameters
        ----------
        mesh_scoping : Scoping, list[int]

        Returns
        -------
        self : Result

        Examples
        --------
        Use a list of nodes.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.complex_rst)
        >>> disp = model.results.displacement
        >>> fc = disp.on_mesh_scoping([1,2,3]).eval()
        >>> len(fc[0].scoping)
        3

        Use a scoping to specify a list of entity IDs with their locations.

        >>> stress = model.results.stress
        >>> scop = dpf.Scoping(ids=[3,4,5], location= dpf.locations.nodal)
        >>> fc = stress.on_mesh_scoping(scop).eval()
        >>> len(fc[0].scoping)
        3
        >>> fc[0].location
        'Nodal'

        """
        if isinstance(mesh_scoping, list):
            mesh_scoping = Scoping(
                ids=mesh_scoping,
                location=self._result_info.native_scoping_location,
                server=self._model._server,
            )

        self._mesh_scoping = mesh_scoping
        return self

    def on_location(self, location):
        """Set the requested location of the provider.

        Elemental nodal fields can be averaged to a nodal or elemental location.

        Parameters
        ----------
        location : str, locations

        Returns
        -------
        self : Result

        Examples
        --------
        Add a requested location to the average result on the nodes.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.complex_rst)
        >>> stress = model.results.stress
        >>> fc = stress.eval()
        >>> fc[0].location
        'ElementalNodal'

        >>> fc = stress.on_location(dpf.locations.nodal).eval()
        >>> fc[0].location
        'Nodal'

        """
        self._location = location
        return self



"""
pow
===
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "math" category
"""

class pow(Operator):
    """Computes element-wise field[i]^p.

      available inputs:
        - field (Field)
        - factor (float)

      available outputs:
        - field (Field)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.math.pow()

      >>> # Make input connections
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_factor = float()
      >>> op.inputs.factor.connect(my_factor)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.math.pow(field=my_field,factor=my_factor)

      >>> # Get output data
      >>> result_field = op.outputs.field()"""
    def __init__(self, field=None, factor=None, config=None, server=None):
        super().__init__(name="Pow", config = config, server = server)
        self._inputs = InputsPow(self)
        self._outputs = OutputsPow(self)
        if field !=None:
            self.inputs.field.connect(field)
        if factor !=None:
            self.inputs.factor.connect(factor)

    @staticmethod
    def _spec():
        spec = Specification(description="""Computes element-wise field[i]^p.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "factor", type_names=["double"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "Pow")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPow 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsPow 
        """
        return super().outputs


#internal name: Pow
#scripting name: pow
class InputsPow(_Inputs):
    """Intermediate class used to connect user inputs to pow operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.pow()
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_factor = float()
      >>> op.inputs.factor.connect(my_factor)
    """
    def __init__(self, op: Operator):
        super().__init__(pow._spec().inputs, op)
        self._field = Input(pow._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._field)
        self._factor = Input(pow._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._factor)

    @property
    def field(self):
        """Allows to connect field input to the operator

        Parameters
        ----------
        my_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.pow()
        >>> op.inputs.field.connect(my_field)
        >>> #or
        >>> op.inputs.field(my_field)

        """
        return self._field

    @property
    def factor(self):
        """Allows to connect factor input to the operator

        Parameters
        ----------
        my_factor : float, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.pow()
        >>> op.inputs.factor.connect(my_factor)
        >>> #or
        >>> op.inputs.factor(my_factor)

        """
        return self._factor

class OutputsPow(_Outputs):
    """Intermediate class used to get outputs from pow operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.pow()
      >>> # Connect inputs : op.inputs. ...
      >>> result_field = op.outputs.field()
    """
    def __init__(self, op: Operator):
        super().__init__(pow._spec().outputs, op)
        self._field = Output(pow._spec().output_pin(0), 0, op) 
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator


        Returns
        ----------
        my_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.pow()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field() 
        """
        return self._field




"""
equivalent_radiated_power
=========================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "result" category
"""

class equivalent_radiated_power(Operator):
    """Compute the Equivalent Radiated Power (ERP)

      available inputs:
        - fields_container (FieldsContainer)
        - meshed_region (MeshedRegion, MeshesContainer) (optional)
        - time_scoping (int, list, Scoping) (optional)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.result.equivalent_radiated_power()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_meshed_region = dpf.MeshedRegion()
      >>> op.inputs.meshed_region.connect(my_meshed_region)
      >>> my_time_scoping = int()
      >>> op.inputs.time_scoping.connect(my_time_scoping)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.result.equivalent_radiated_power(fields_container=my_fields_container,meshed_region=my_meshed_region,time_scoping=my_time_scoping)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, fields_container=None, meshed_region=None, time_scoping=None, config=None, server=None):
        super().__init__(name="ERP", config = config, server = server)
        self._inputs = InputsEquivalentRadiatedPower(self)
        self._outputs = OutputsEquivalentRadiatedPower(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)
        if meshed_region !=None:
            self.inputs.meshed_region.connect(meshed_region)
        if time_scoping !=None:
            self.inputs.time_scoping.connect(time_scoping)

    @staticmethod
    def _spec():
        spec = Specification(description="""Compute the Equivalent Radiated Power (ERP)""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "meshed_region", type_names=["abstract_meshed_region","meshes_container"], optional=True, document="""the mesh region in this pin have to be boundary or skin mesh"""), 
                                 2 : PinSpecification(name = "time_scoping", type_names=["int32","vector<int32>","scoping"], optional=True, document="""load step number (if it's specified, the ERP is computed only on the substeps of this step) or time scoping""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "ERP")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsEquivalentRadiatedPower 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsEquivalentRadiatedPower 
        """
        return super().outputs


#internal name: ERP
#scripting name: equivalent_radiated_power
class InputsEquivalentRadiatedPower(_Inputs):
    """Intermediate class used to connect user inputs to equivalent_radiated_power operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.equivalent_radiated_power()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_meshed_region = dpf.MeshedRegion()
      >>> op.inputs.meshed_region.connect(my_meshed_region)
      >>> my_time_scoping = int()
      >>> op.inputs.time_scoping.connect(my_time_scoping)
    """
    def __init__(self, op: Operator):
        super().__init__(equivalent_radiated_power._spec().inputs, op)
        self._fields_container = Input(equivalent_radiated_power._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)
        self._meshed_region = Input(equivalent_radiated_power._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._meshed_region)
        self._time_scoping = Input(equivalent_radiated_power._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._time_scoping)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.equivalent_radiated_power()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

    @property
    def meshed_region(self):
        """Allows to connect meshed_region input to the operator

        - pindoc: the mesh region in this pin have to be boundary or skin mesh

        Parameters
        ----------
        my_meshed_region : MeshedRegion, MeshesContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.equivalent_radiated_power()
        >>> op.inputs.meshed_region.connect(my_meshed_region)
        >>> #or
        >>> op.inputs.meshed_region(my_meshed_region)

        """
        return self._meshed_region

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator

        - pindoc: load step number (if it's specified, the ERP is computed only on the substeps of this step) or time scoping

        Parameters
        ----------
        my_time_scoping : int, list, Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.equivalent_radiated_power()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> #or
        >>> op.inputs.time_scoping(my_time_scoping)

        """
        return self._time_scoping

class OutputsEquivalentRadiatedPower(_Outputs):
    """Intermediate class used to get outputs from equivalent_radiated_power operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.equivalent_radiated_power()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(equivalent_radiated_power._spec().outputs, op)
        self._fields_container = Output(equivalent_radiated_power._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.equivalent_radiated_power()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container

